# copyright 2003 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.server.migractions"""

import os
import os.path as osp
import sys
from contextlib import contextmanager
from datetime import date
from tempfile import TemporaryDirectory

from logilab.common import tempattr
from yams.constraints import UniqueConstraint

from yams import ValidationError
from cubicweb import ConfigurationError, Binary
from cubicweb.devtools import startpgcluster, stoppgcluster
from cubicweb.devtools.apptest_config import PostgresApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.schema import constraint_name_for
from cubicweb.server.migractions import ServerMigrationHelper
from cubicweb.server.schema2sql import build_index_name
from cubicweb.server.sources import storages
from cubicweb.server.sqlutils import SQL_PREFIX

HERE = osp.dirname(osp.abspath(__file__))
migrschema = None


def setUpModule():
    startpgcluster(__file__)


def tearDownModule(*args):
    global migrschema
    del migrschema
    if hasattr(MigrationCommandsTC, "origschema"):
        del MigrationCommandsTC.origschema
    stoppgcluster(__file__)


class MigrationTC(CubicWebTC):
    appid = "data-migractions"

    configcls = PostgresApptestConfiguration

    def _init_repo(self):
        super()._init_repo()
        # we have to read schema from the database to get eid for schema entities
        self.repo.set_schema(self.repo.deserialize_schema(), resetvreg=False)
        # hack to read the schema from data/migrschema

        @contextmanager
        def temp_app(config, appid, apphome):
            old = config.apphome, config.appid
            sys.path.remove(old[0])
            sys.path.insert(0, apphome)
            config._apphome, config.appid = apphome, appid
            try:
                yield config
            finally:
                sys.path.remove(apphome)
                sys.path.insert(0, old[0])
                config._apphome, config.appid = old

        appid = osp.join(self.appid, "migratedapp")
        apphome = osp.join(HERE, appid)
        with temp_app(self.config, appid, apphome) as config:
            global migrschema
            migrschema = config.load_schema()

    def tearDown(self):
        super().tearDown()
        self.repo.vreg["etypes"].clear_caches()

    @contextmanager
    def mh(self):
        with self.admin_access.repo_cnx() as cnx:
            yield cnx, ServerMigrationHelper(
                self.repo.config, migrschema, repo=self.repo, cnx=cnx, interactive=False
            )

    def table_sql(self, mh, tablename):
        result = mh.sqlexec(
            "SELECT table_name FROM information_schema.tables WHERE LOWER(table_name)=%(table)s",
            {"table": tablename.lower()},
        )
        if result:
            return result[0][0]
        return None  # no such table

    def table_schema(self, mh, tablename):
        result = mh.sqlexec(
            "SELECT column_name, data_type, character_maximum_length "
            "FROM information_schema.columns "
            "WHERE LOWER(table_name) = %(table)s",
            {"table": tablename.lower()},
        )
        assert result, f"no table {tablename}"
        return {x[0]: (x[1], x[2]) for x in result}

    def table_constraints(self, mh, tablename):
        result = mh.sqlexec(
            "SELECT DISTINCT constraint_name FROM information_schema.constraint_column_usage "
            "WHERE LOWER(table_name) = '%(table)s' AND constraint_name LIKE 'cstr%%'"
            % {"table": tablename.lower()}
        )
        assert result, f"no table {tablename}"
        return {x[0] for x in result}


class MigrationCommandsTC(MigrationTC):
    def _init_repo(self):
        super()._init_repo()
        assert "Folder" in migrschema

    def test_add_attribute_bool(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn("yesno", self.schema)
            cnx.create_entity("Note")
            cnx.commit()
            mh.cmd_add_attribute("Note", "yesno")
            self.assertIn("yesno", self.schema)
            self.assertEqual(self.schema["yesno"].subjects(), ("Note",))
            self.assertEqual(self.schema["yesno"].objects(), ("Boolean",))
            self.assertEqual(self.schema["Note"].default("yesno"), False)
            # test default value set on existing entities
            note = cnx.execute("Note X").get_entity(0, 0)
            self.assertEqual(note.yesno, False)
            # test default value set for next entities
            self.assertEqual(cnx.create_entity("Note").yesno, False)

    def test_add_attribute_int(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn("whatever", self.schema)
            cnx.create_entity("Note")
            cnx.commit()
            orderdict = dict(
                mh.rqlexec(
                    'Any RTN, O WHERE X name "Note", RDEF from_entity X, '
                    "RDEF relation_type RT, RDEF ordernum O, RT name RTN"
                )
            )
            mh.cmd_add_attribute("Note", "whatever")
            self.assertIn("whatever", self.schema)
            self.assertEqual(self.schema["whatever"].subjects(), ("Note",))
            self.assertEqual(self.schema["whatever"].objects(), ("Int",))
            self.assertEqual(self.schema["Note"].default("whatever"), 0)
            # test default value set on existing entities
            note = cnx.execute("Note X").get_entity(0, 0)
            self.assertIsInstance(note.whatever, int)
            self.assertEqual(note.whatever, 0)
            # test default value set for next entities
            self.assertEqual(cnx.create_entity("Note").whatever, 0)
            # test attribute order
            orderdict2 = dict(
                mh.rqlexec(
                    'Any RTN, O WHERE X name "Note", RDEF from_entity X, '
                    "RDEF relation_type RT, RDEF ordernum O, RT name RTN"
                )
            )
            whateverorder = (
                migrschema["whatever"].relation_definition("Note", "Int").order
            )
            for k, v in orderdict.items():
                if v >= whateverorder:
                    orderdict[k] = v + 1
            orderdict["whatever"] = whateverorder
            self.assertDictEqual(orderdict, orderdict2)

    def test_add_attribute_varchar(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn("whatever", self.schema)
            cnx.create_entity("Note")
            cnx.commit()
            self.assertNotIn("shortpara", self.schema)
            mh.cmd_add_attribute("Note", "shortpara")
            self.assertIn("shortpara", self.schema)
            self.assertEqual(self.schema["shortpara"].subjects(), ("Note",))
            self.assertEqual(self.schema["shortpara"].objects(), ("String",))
            # test created column is actually a varchar(64)
            fields = self.table_schema(mh, f"{SQL_PREFIX}Note")
            self.assertEqual(
                fields[f"{SQL_PREFIX}shortpara"], ("character varying", 11)
            )
            # test default value set on existing entities
            self.assertEqual(cnx.execute("Note X").get_entity(0, 0).shortpara, "hop")
            # test default value set for next entities
            self.assertEqual(
                cnx.create_entity("Note", shortpara="hop hop").shortpara, "hop hop"
            )
            # serialized constraint added
            constraints = self.table_constraints(mh, "cw_Personne")
            self.assertEqual(len(constraints), 1, constraints)

    def test_add_attribute_unique(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn("unique_id", self.schema)
            mh.cmd_add_attribute("Note", "unique_id")
            # test unique index creation
            dbh = self.repo.system_source.dbhelper
            self.assertTrue(
                dbh.index_exists(cnx.cnxset.cu, "cw_Note", "cw_unique_id", unique=True)
            )

    def test_add_datetime_with_default_value_attribute(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn("mydate", self.schema)
            self.assertNotIn("oldstyledefaultdate", self.schema)
            self.assertNotIn("newstyledefaultdate", self.schema)
            mh.cmd_add_attribute("Note", "mydate")
            mh.cmd_add_attribute("Note", "oldstyledefaultdate")
            mh.cmd_add_attribute("Note", "newstyledefaultdate")
            self.assertIn("mydate", self.schema)
            self.assertIn("oldstyledefaultdate", self.schema)
            self.assertIn("newstyledefaultdate", self.schema)
            self.assertEqual(self.schema["mydate"].subjects(), ("Note",))
            self.assertEqual(self.schema["mydate"].objects(), ("Date",))
            testdate = date(2005, 12, 13)
            eid1 = mh.rqlexec("INSERT Note N")[0][0]
            eid2 = mh.rqlexec(
                "INSERT Note N: N mydate %(mydate)s", {"mydate": testdate}
            )[0][0]
            d1 = mh.rqlexec("Any D WHERE X eid %(x)s, X mydate D", {"x": eid1})[0][0]
            d2 = mh.rqlexec("Any D WHERE X eid %(x)s, X mydate D", {"x": eid2})[0][0]
            d3 = mh.rqlexec(
                "Any D WHERE X eid %(x)s, X oldstyledefaultdate D", {"x": eid1}
            )[0][0]
            d4 = mh.rqlexec(
                "Any D WHERE X eid %(x)s, X newstyledefaultdate D", {"x": eid1}
            )[0][0]
            self.assertEqual(d1, date.today())
            self.assertEqual(d2, testdate)
            myfavoritedate = date(2013, 1, 1)
            self.assertEqual(d3, myfavoritedate)
            self.assertEqual(d4, myfavoritedate)

    def test_drop_chosen_constraints_ctxmanager(self):
        with self.mh() as (cnx, mh):
            with mh.cmd_dropped_constraints("Note", "unique_id", UniqueConstraint):
                mh.cmd_add_attribute("Note", "unique_id")
                # make sure the maxsize constraint is not dropped
                self.assertRaises(
                    ValidationError, mh.rqlexec, 'INSERT Note N: N unique_id "xyz"'
                )
                mh.rollback()
                # make sure the unique constraint is dropped
                mh.rqlexec('INSERT Note N: N unique_id "x"')
                mh.rqlexec('INSERT Note N: N unique_id "x"')
                mh.rqlexec("DELETE Note N")

    def test_drop_required_ctxmanager(self):
        with self.mh() as (cnx, mh):
            with mh.cmd_dropped_constraints(
                "Note", "unique_id", cstrtype=None, droprequired=True
            ):
                mh.cmd_add_attribute("Note", "unique_id")
                mh.rqlexec("INSERT Note N")
                mh.rqlexec('SET N unique_id "x"')
            # make sure the required=True was restored
            self.assertRaises(ValidationError, mh.rqlexec, "INSERT Note N")
            mh.rollback()

    def test_rename_attribute(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn("civility", self.schema)
            eid1 = mh.rqlexec('INSERT Personne X: X nom "lui", X sexe "M"')[0][0]
            eid2 = mh.rqlexec('INSERT Personne X: X nom "l\'autre", X sexe NULL')[0][0]
            mh.cmd_rename_attribute("Personne", "sexe", "civility")
            self.assertNotIn("sexe", self.schema)
            self.assertIn("civility", self.schema)
            # test data has been backported
            c1 = mh.rqlexec(f"Any C WHERE X eid {eid1}, X civility C")[0][0]
            self.assertEqual(c1, "M")
            c2 = mh.rqlexec(f"Any C WHERE X eid {eid2}, X civility C")[0][0]
            self.assertEqual(c2, None)

    def test_workflow_actions(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_workflow("foo", ("Personne", "Email"), ensure_workflowable=False)
            for etype in ("Personne", "Email"):
                s1 = mh.rqlexec(
                    f'Any N WHERE WF workflow_of ET, ET name "{etype}", WF name N'
                )[0][0]
                self.assertEqual(s1, "foo")
                s1 = mh.rqlexec(
                    f'Any N WHERE ET default_workflow WF, ET name "{etype}", WF name N'
                )[0][0]
                self.assertEqual(s1, "foo")

    def test_add_entity_type(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn("Folder2", self.schema)
            self.assertNotIn("filed_under2", self.schema)
            mh.cmd_add_entity_type("Folder2")
            self.assertIn("Folder2", self.schema)
            self.assertIn("Old", self.schema)
            self.assertTrue(cnx.execute('CWEType X WHERE X name "Folder2"'))
            self.assertIn("filed_under2", self.schema)
            self.assertTrue(cnx.execute('CWRType X WHERE X name "filed_under2"'))
            self.assertEqual(
                sorted(self.schema["Folder2"].subject_relations),
                [
                    "created_by",
                    "creation_date",
                    "cw_source",
                    "cwuri",
                    "description",
                    "description_format",
                    "eid",
                    "filed_under2",
                    "has_text",
                    "identity",
                    "in_basket",
                    "inlined_rel",
                    "is",
                    "is_instance_of",
                    "modification_date",
                    "name",
                    "owned_by",
                ],
            )
            self.assertCountEqual(
                [str(rs) for rs in self.schema["Folder2"].object_relations.values()],
                ["filed_under2", "identity", "inlined_rel"],
            )
            # Old will be missing as it has been renamed into 'New' in the migrated
            # schema while New hasn't been added here.
            self.assertEqual(
                sorted(str(e) for e in self.schema["filed_under2"].subjects()),
                sorted(
                    str(e) for e in self.schema.entities() if not e.final and e != "Old"
                ),
            )
            self.assertEqual(self.schema["filed_under2"].objects(), ("Folder2",))
            eschema = self.schema.entity_schema_for("Folder2")
            for cstr in eschema.relation_definition("name").constraints:
                self.assertTrue(hasattr(cstr, "eid"))

    def test_add_entity_type_with_constraint(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_entity_type("Activity")
            constraints = self.table_constraints(mh, "cw_Activity")
            self.assertEqual(len(constraints), 2, constraints)

    def test_add_cube_with_custom_final_type(self):
        with self.mh() as (cnx, mh):
            try:
                mh.cmd_add_cube("fakecustomtype")
                self.assertIn("Numeric", self.schema)
                self.assertTrue(self.schema["Numeric"].final)
                rdef = self.schema["num"].relation_definitions[("Location", "Numeric")]
                self.assertEqual(rdef.scale, 10)
                self.assertEqual(rdef.precision, 18)
                fields = self.table_schema(mh, f"{SQL_PREFIX}Location")
                self.assertEqual(fields[f"{SQL_PREFIX}num"], ("numeric", None))  # XXX
            finally:
                mh.cmd_drop_cube("fakecustomtype")
                mh.drop_entity_type("Numeric")

    def test_add_drop_entity_type(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_entity_type("Folder2")
            wf = mh.cmd_add_workflow("folder2 wf", "Folder2", ensure_workflowable=False)
            todo = wf.add_state("todo", initial=True)
            done = wf.add_state("done")
            wf.add_transition("redoit", done, todo)
            wf.add_transition("markasdone", todo, done)
            cnx.commit()
            mh.cmd_drop_entity_type("Folder2")
            self.assertNotIn("Folder2", self.schema)
            self.assertFalse(cnx.execute('CWEType X WHERE X name "Folder2"'))
            # test automatic workflow deletion
            self.assertFalse(cnx.execute("Workflow X WHERE NOT X workflow_of ET"))
            self.assertFalse(cnx.execute("State X WHERE NOT X state_of WF"))
            self.assertFalse(cnx.execute("Transition X WHERE NOT X transition_of WF"))

    def test_rename_entity_type(self):
        with self.mh() as (cnx, mh):
            entity = mh.create_entity("Old", name="old")
            self.repo.type_from_eid(entity.eid, entity._cw)
            mh.cmd_rename_entity_type("Old", "New")
            dbh = self.repo.system_source.dbhelper
            indices = set(dbh.list_indices(cnx.cnxset.cu, "cw_New"))
            self.assertNotIn(build_index_name("cw_Old", ["cw_name"]), indices)
            self.assertIn(build_index_name("cw_New", ["cw_name"]), indices)
            mh.cmd_rename_attribute("New", "name", "new_name")
            indices = set(dbh.list_indices(cnx.cnxset.cu, "cw_New"))
            # check 'indexed' index
            self.assertNotIn(build_index_name("cw_New", ["cw_name"], "idx_"), indices)
            self.assertIn(build_index_name("cw_New", ["cw_new_name"], "idx_"), indices)
            # check 'unique' index
            self.assertNotIn(build_index_name("cw_New", ["cw_name"], "key_"), indices)
            self.assertIn(build_index_name("cw_New", ["cw_new_name"], "key_"), indices)

    def test_add_drop_relation_type(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_entity_type("Folder2", auto=False)
            mh.cmd_add_relation_type("filed_under2")
            self.assertIn("filed_under2", self.schema)
            # Old will be missing as it has been renamed into 'New' in the migrated
            # schema while New hasn't been added here.
            self.assertEqual(
                sorted(str(e) for e in self.schema["filed_under2"].subjects()),
                sorted(
                    str(e) for e in self.schema.entities() if not e.final and e != "Old"
                ),
            )
            self.assertEqual(self.schema["filed_under2"].objects(), ("Folder2",))
            mh.cmd_drop_relation_type("filed_under2")
            self.assertNotIn("filed_under2", self.schema)
            # this should not crash
            mh.cmd_drop_relation_type("filed_under2")

    def test_add_relation_definition_nortype(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_relation_definition("Personne", "concerne2", "Affaire")
            self.assertEqual(self.schema["concerne2"].subjects(), ("Personne",))
            self.assertEqual(self.schema["concerne2"].objects(), ("Affaire",))
            self.assertEqual(
                self.schema["concerne2"]
                .relation_definition("Personne", "Affaire")
                .cardinality,
                "1*",
            )
            mh.cmd_add_relation_definition("Personne", "concerne2", "Note")
            self.assertEqual(
                sorted(self.schema["concerne2"].objects()), ["Affaire", "Note"]
            )
            mh.create_entity("Personne", nom="tot")
            mh.create_entity("Affaire")
            mh.rqlexec("SET X concerne2 Y WHERE X is Personne, Y is Affaire")
            cnx.commit()
            mh.cmd_drop_relation_definition("Personne", "concerne2", "Affaire")
            self.assertIn("concerne2", self.schema)
            mh.cmd_drop_relation_definition("Personne", "concerne2", "Note")
            self.assertNotIn("concerne2", self.schema)

    def test_drop_relation_definition_existant_rtype(self):
        with self.mh() as (cnx, mh):
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].subjects()),
                ["Affaire", "Personne"],
            )
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].objects()),
                ["Affaire", "Division", "Note", "Societe", "SubDivision"],
            )
            mh.cmd_drop_relation_definition("Personne", "concerne", "Affaire")
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].subjects()), ["Affaire"]
            )
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].objects()),
                ["Division", "Note", "Societe", "SubDivision"],
            )
            mh.cmd_add_relation_definition("Personne", "concerne", "Affaire")
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].subjects()),
                ["Affaire", "Personne"],
            )
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].objects()),
                ["Affaire", "Division", "Note", "Societe", "SubDivision"],
            )
            # trick: overwrite self.maxeid to avoid deletion of just reintroduced types
            self.maxeid = cnx.execute("Any MAX(X)")[0][0]

    def test_drop_relation_definition_with_specialization(self):
        with self.mh() as (cnx, mh):
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].subjects()),
                ["Affaire", "Personne"],
            )
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].objects()),
                ["Affaire", "Division", "Note", "Societe", "SubDivision"],
            )
            mh.cmd_drop_relation_definition("Affaire", "concerne", "Societe")
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].subjects()),
                ["Affaire", "Personne"],
            )
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].objects()),
                ["Affaire", "Note"],
            )
            mh.cmd_add_relation_definition("Affaire", "concerne", "Societe")
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].subjects()),
                ["Affaire", "Personne"],
            )
            self.assertEqual(
                sorted(str(e) for e in self.schema["concerne"].objects()),
                ["Affaire", "Division", "Note", "Societe", "SubDivision"],
            )
            # trick: overwrite self.maxeid to avoid deletion of just reintroduced types
            self.maxeid = cnx.execute("Any MAX(X)")[0][0]

    def test_rename_relation(self):
        self.skipTest("implement me")

    def test_change_relation_props_non_final(self):
        with self.mh() as (cnx, mh):
            rschema = self.schema["concerne"]
            card = rschema.relation_definition("Affaire", "Societe").cardinality
            self.assertEqual(card, "**")
            try:
                mh.cmd_change_relation_props(
                    "Affaire", "concerne", "Societe", cardinality="?*"
                )
                card = rschema.relation_definition("Affaire", "Societe").cardinality
                self.assertEqual(card, "?*")
            finally:
                mh.cmd_change_relation_props(
                    "Affaire", "concerne", "Societe", cardinality="**"
                )

    def test_change_relation_props_final(self):
        with self.mh() as (cnx, mh):
            rschema = self.schema["adel"]
            card = rschema.relation_definition("Personne", "String").fulltextindexed
            self.assertEqual(card, False)
            try:
                mh.cmd_change_relation_props(
                    "Personne", "adel", "String", fulltextindexed=True
                )
                card = rschema.relation_definition("Personne", "String").fulltextindexed
                self.assertEqual(card, True)
            finally:
                mh.cmd_change_relation_props(
                    "Personne", "adel", "String", fulltextindexed=False
                )

    def test_sync_schema_props_perms_rqlconstraints(self):
        with self.mh() as (cnx, mh):
            # Drop one of the RQLConstraint.
            rdef = self.schema["evaluee"].relation_definitions[("Personne", "Note")]
            oldconstraints = rdef.constraints
            self.assertIn(
                "S created_by U", [cstr.expression for cstr in oldconstraints]
            )
            mh.cmd_sync_schema_props_perms("evaluee", commit=True)
            newconstraints = rdef.constraints
            self.assertNotIn(
                "S created_by U", [cstr.expression for cstr in newconstraints]
            )

            # Drop all RQLConstraint.
            rdef = self.schema["travaille"].relation_definitions[
                ("Personne", "Societe")
            ]
            oldconstraints = rdef.constraints
            self.assertEqual(len(oldconstraints), 2)
            mh.cmd_sync_schema_props_perms("travaille", commit=True)
            rdef = self.schema["travaille"].relation_definitions[
                ("Personne", "Societe")
            ]
            newconstraints = rdef.constraints
            self.assertEqual(len(newconstraints), 0)

    def test_sync_schema_props_perms(self):
        with self.mh() as (cnx, mh):
            nbrqlexpr_start = cnx.execute("Any COUNT(X) WHERE X is RQLExpression")[0][0]
            migrschema["titre"].relation_definitions[("Personne", "String")].order = 7
            migrschema["adel"].relation_definitions[("Personne", "String")].order = 6
            migrschema["ass"].relation_definitions[("Personne", "String")].order = 5
            migrschema["Personne"].description = "blabla bla"
            migrschema["titre"].description = "usually a title"
            migrschema["titre"].relation_definitions[
                ("Personne", "String")
            ].description = "title for this person"
            delete_concerne_rqlexpr = self._rrqlexpr_rset(cnx, "delete", "concerne")
            add_concerne_rqlexpr = self._rrqlexpr_rset(cnx, "add", "concerne")

            # make sure properties (e.g. etype descriptions) are synced by the
            # second call to sync_schema
            mh.cmd_sync_schema_props_perms(syncprops=False, commit=False)
            mh.cmd_sync_schema_props_perms(commit=False)

            self.assertEqual(
                cnx.execute('Any D WHERE X name "Personne", X description D')[0][0],
                "blabla bla",
            )
            self.assertEqual(
                cnx.execute('Any D WHERE X name "titre", X description D')[0][0],
                "usually a title",
            )
            self.assertEqual(
                cnx.execute(
                    'Any D WHERE X relation_type RT, RT name "titre",'
                    'X from_entity FE, FE name "Personne",'
                    "X description D"
                )[0][0],
                "title for this person",
            )
            rinorder = [
                n
                for n, in cnx.execute(
                    "Any N ORDERBY O,N WHERE X is CWAttribute, X relation_type RT, RT name N,"
                    'X from_entity FE, FE name "Personne",'
                    "X ordernum O"
                )
            ]
            expected = [
                "nom",
                "prenom",
                "sexe",
                "promo",
                "ass",
                "adel",
                "titre",
                "web",
                "tel",
                "fax",
                "datenaiss",
                "test",
                "tzdatenaiss",
                "description",
                "firstname",
                "photo",
                "creation_date",
                "cwuri",
                "modification_date",
            ]
            self.assertEqual(expected, rinorder)

            # test permissions synchronization ####################################
            # new rql expr to add note entity
            eexpr = self._erqlexpr_entity(cnx, "add", "Note")
            self.assertEqual(
                eexpr.expression,
                "X ecrit_part PE, U in_group G, "
                'PE require_permission P, P name "add_note", P require_group G',
            )
            self.assertEqual([et.name for et in eexpr.reverse_add_permission], ["Note"])
            self.assertEqual(eexpr.reverse_read_permission, ())
            self.assertEqual(eexpr.reverse_delete_permission, ())
            self.assertEqual(eexpr.reverse_update_permission, ())
            self.assertTrue(self._rrqlexpr_rset(cnx, "add", "para"))
            # no rqlexpr to delete para attribute
            self.assertFalse(self._rrqlexpr_rset(cnx, "delete", "para"))
            # new rql expr to add ecrit_par relation
            rexpr = self._rrqlexpr_entity(cnx, "add", "ecrit_par")
            self.assertEqual(
                rexpr.expression,
                'O require_permission P, P name "add_note", '
                "U in_group G, P require_group G",
            )
            self.assertEqual(
                [rdef.relation_type[0].name for rdef in rexpr.reverse_add_permission],
                ["ecrit_par"],
            )
            self.assertEqual(rexpr.reverse_read_permission, ())
            self.assertEqual(rexpr.reverse_delete_permission, ())
            # no more rqlexpr to delete and add travaille relation
            self.assertFalse(self._rrqlexpr_rset(cnx, "add", "travaille"))
            self.assertFalse(self._rrqlexpr_rset(cnx, "delete", "travaille"))
            # no more rqlexpr to delete and update Societe entity
            self.assertFalse(self._erqlexpr_rset(cnx, "update", "Societe"))
            self.assertFalse(self._erqlexpr_rset(cnx, "delete", "Societe"))
            # no more rqlexpr to read Affaire entity
            self.assertFalse(self._erqlexpr_rset(cnx, "read", "Affaire"))
            # rqlexpr to update Affaire entity has been updated
            eexpr = self._erqlexpr_entity(cnx, "update", "Affaire")
            self.assertEqual(eexpr.expression, "X concerne S, S owned_by U")
            # no change for rqlexpr to add and delete Affaire entity
            self.assertEqual(len(self._erqlexpr_rset(cnx, "delete", "Affaire")), 1)
            self.assertEqual(len(self._erqlexpr_rset(cnx, "add", "Affaire")), 1)
            # no change for rqlexpr to add and delete concerne relation
            self.assertEqual(
                len(self._rrqlexpr_rset(cnx, "delete", "concerne")),
                len(delete_concerne_rqlexpr),
            )
            self.assertEqual(
                len(self._rrqlexpr_rset(cnx, "add", "concerne")),
                len(add_concerne_rqlexpr),
            )
            # * migrschema involve:
            #   * 7 erqlexprs deletions (2 in (Affaire + Societe + Note.para) + 1 Note.something
            #   * 2 rrqlexprs deletions (travaille)
            #   * 1 update (Affaire update)
            #   * 2 new (Note add, ecrit_par add)
            #   * 2 implicit new for attributes (Note.para, Person.test)
            # remaining orphan rql expr which should be deleted at commit (composite relation)
            # unattached expressions -> pending deletion on commit
            self.assertEqual(
                cnx.execute(
                    'Any COUNT(X) WHERE X is RQLExpression, X exprtype "ERQLExpression",'
                    "NOT ET1 read_permission X, NOT ET2 add_permission X, "
                    "NOT ET3 delete_permission X, NOT ET4 update_permission X"
                )[0][0],
                7,
            )
            self.assertEqual(
                cnx.execute(
                    'Any COUNT(X) WHERE X is RQLExpression, X exprtype "RRQLExpression",'
                    "NOT ET1 read_permission X, NOT ET2 add_permission X, "
                    "NOT ET3 delete_permission X, NOT ET4 update_permission X"
                )[0][0],
                2,
            )
            # finally
            self.assertEqual(
                cnx.execute("Any COUNT(X) WHERE X is RQLExpression")[0][0],
                nbrqlexpr_start + 1 + 2 + 2 + 2,
            )
            cnx.commit()
            # unique_together test
            self.assertEqual(
                len(self.schema.entity_schema_for("Personne")._unique_together), 1
            )
            self.assertCountEqual(
                self.schema.entity_schema_for("Personne")._unique_together[0],
                ("nom", "prenom", "datenaiss"),
            )
            rset = cnx.execute(
                "Any C WHERE C is CWUniqueTogetherConstraint, "
                'C constraint_of ET, ET name "Personne"'
            )
            self.assertEqual(len(rset), 1)
            relations = [r.name for r in rset.get_entity(0, 0).relations]
            self.assertCountEqual(relations, ("nom", "prenom", "datenaiss"))

            # serialized constraint changed
            constraints = self.table_constraints(mh, "cw_Personne")
            self.assertEqual(len(constraints), 1, constraints)
            rdef = migrschema["promo"].relation_definitions["Personne", "String"]
            cstr = rdef.constraint_by_type("StaticVocabularyConstraint")
            self.assertIn(constraint_name_for(cstr, rdef), constraints)

    def _erqlexpr_rset(self, cnx, action, ertype):
        rql = (
            "RQLExpression X WHERE ET is CWEType, ET %s_permission X, ET name %%(name)s"
            % action
        )
        return cnx.execute(rql, {"name": ertype})

    def _erqlexpr_entity(self, cnx, action, ertype):
        rset = self._erqlexpr_rset(cnx, action, ertype)
        self.assertEqual(len(rset), 1)
        return rset.get_entity(0, 0)

    def _rrqlexpr_rset(self, cnx, action, ertype):
        return cnx.execute(
            "RQLExpression X WHERE RT is CWRType, RDEF %s_permission X, "
            "RT name %%(name)s, RDEF relation_type RT" % action,
            {"name": ertype},
        )

    def _rrqlexpr_entity(self, cnx, action, ertype):
        rset = self._rrqlexpr_rset(cnx, action, ertype)
        self.assertEqual(len(rset), 1)
        return rset.get_entity(0, 0)

    def test_set_size_constraint(self):
        with self.mh() as (cnx, mh):
            # existing previous value
            try:
                mh.cmd_set_size_constraint("CWEType", "name", 128)
            finally:
                mh.cmd_set_size_constraint("CWEType", "name", 64)
            # non existing previous value
            try:
                mh.cmd_set_size_constraint("CWEType", "description", 256)
            finally:
                mh.cmd_set_size_constraint("CWEType", "description", None)

    def test_add_drop_cube_and_deps(self):
        with self.mh() as (cnx, mh):
            schema = self.repo.schema
            self.assertEqual(
                sorted(
                    (str(s), str(o)) for s, o in schema["see_also"].relation_definitions
                ),
                sorted(
                    [
                        ("EmailThread", "EmailThread"),
                        ("Folder", "Folder"),
                        ("Bookmark", "Bookmark"),
                        ("Bookmark", "Note"),
                        ("Note", "Note"),
                        ("Note", "Bookmark"),
                    ]
                ),
            )
            try:
                mh.cmd_drop_cube("fakeemail", removedeps=True)
                # file was there because it's an email dependancy, should have been removed
                self.assertNotIn("fakeemail", self.config.cubes())
                self.assertNotIn(
                    self.config.cube_dir("fakeemail"), self.config.cubes_path()
                )
                self.assertNotIn("file", self.config.cubes())
                self.assertNotIn(self.config.cube_dir("file"), self.config.cubes_path())
                for ertype in (
                    "Email",
                    "EmailThread",
                    "EmailPart",
                    "File",
                    "sender",
                    "in_thread",
                    "reply_to",
                    "data_format",
                ):
                    self.assertNotIn(ertype, schema)
                self.assertEqual(
                    sorted(schema["see_also"].relation_definitions),
                    sorted(
                        [
                            ("Folder", "Folder"),
                            ("Bookmark", "Bookmark"),
                            ("Bookmark", "Note"),
                            ("Note", "Note"),
                            ("Note", "Bookmark"),
                        ]
                    ),
                )
                self.assertEqual(
                    sorted(schema["see_also"].subjects()),
                    ["Bookmark", "Folder", "Note"],
                )
                self.assertEqual(
                    sorted(schema["see_also"].objects()), ["Bookmark", "Folder", "Note"]
                )
                self.assertEqual(
                    cnx.execute(
                        'Any X WHERE X pkey "system.version.fakeemail"'
                    ).rowcount,
                    0,
                )
                self.assertEqual(
                    cnx.execute('Any X WHERE X pkey "system.version.file"').rowcount, 0
                )
            finally:
                mh.cmd_add_cube("fakeemail")
                self.assertIn("fakeemail", self.config.cubes())
                self.assertIn(
                    self.config.cube_dir("fakeemail"), self.config.cubes_path()
                )
                self.assertIn("file", self.config.cubes())
                self.assertIn(self.config.cube_dir("file"), self.config.cubes_path())
                for ertype in (
                    "Email",
                    "EmailThread",
                    "EmailPart",
                    "File",
                    "sender",
                    "in_thread",
                    "reply_to",
                    "data_format",
                ):
                    self.assertIn(ertype, schema)
                self.assertEqual(
                    sorted(schema["see_also"].relation_definitions),
                    sorted(
                        [
                            ("EmailThread", "EmailThread"),
                            ("Folder", "Folder"),
                            ("Bookmark", "Bookmark"),
                            ("Bookmark", "Note"),
                            ("Note", "Note"),
                            ("Note", "Bookmark"),
                        ]
                    ),
                )
                self.assertEqual(
                    sorted(schema["see_also"].subjects()),
                    ["Bookmark", "EmailThread", "Folder", "Note"],
                )
                self.assertEqual(
                    sorted(schema["see_also"].objects()),
                    ["Bookmark", "EmailThread", "Folder", "Note"],
                )
                from cubicweb_fakeemail.__pkginfo__ import version as email_version
                from cubicweb_file.__pkginfo__ import version as file_version

                self.assertEqual(
                    cnx.execute(
                        'Any V WHERE X value V, X pkey "system.version.fakeemail"'
                    )[0][0],
                    email_version,
                )
                self.assertEqual(
                    cnx.execute('Any V WHERE X value V, X pkey "system.version.file"')[
                        0
                    ][0],
                    file_version,
                )
                # why this commit is necessary is unclear to me (though without it
                # next test may fail complaining of missing tables
                cnx.commit()

    def test_add_drop_cube_no_deps(self):
        with self.mh() as (cnx, mh):
            cubes = set(self.config.cubes())
            schema = self.repo.schema
            try:
                mh.cmd_drop_cube("fakeemail")
                cubes.remove("fakeemail")
                self.assertNotIn("fakeemail", self.config.cubes())
                self.assertIn("file", self.config.cubes())
                for ertype in (
                    "Email",
                    "EmailThread",
                    "EmailPart",
                    "sender",
                    "in_thread",
                    "reply_to",
                ):
                    self.assertNotIn(ertype, schema)
            finally:
                mh.cmd_add_cube("fakeemail")
                self.assertIn("fakeemail", self.config.cubes())
                # why this commit is necessary is unclear to me (though without it
                # next test may fail complaining of missing tables
                cnx.commit()

    def test_drop_dep_cube(self):
        with self.mh() as (cnx, mh):
            with self.assertRaises(ConfigurationError) as cm:
                mh.cmd_drop_cube("file")
            self.assertEqual(
                str(cm.exception),
                "can't remove cube file, used as a dependency."
                "Consider using 'force_drop=True' if you want to drop it anyway",
            )

    def test_introduce_base_class(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_entity_type("Para")
            self.assertEqual(
                sorted(et.type for et in self.schema["Para"].specialized_by()), ["Note"]
            )
            self.assertEqual(self.schema["Note"].specializes().type, "Para")
            mh.cmd_add_entity_type("Text")
            self.assertEqual(
                sorted(et.type for et in self.schema["Para"].specialized_by()),
                ["Note", "Text"],
            )
            self.assertEqual(self.schema["Text"].specializes().type, "Para")
            # test columns have been actually added
            text = cnx.create_entity("Text", para="hip", summary="hop", newattr="momo")
            note = cnx.create_entity(
                "Note", para="hip", shortpara="hop", newattr="momo", unique_id="x"
            )
            aff = cnx.create_entity("Affaire")
            self.assertTrue(
                cnx.execute(
                    "SET X newnotinlined Y WHERE X eid %(x)s, Y eid %(y)s",
                    {"x": text.eid, "y": aff.eid},
                )
            )
            self.assertTrue(
                cnx.execute(
                    "SET X newnotinlined Y WHERE X eid %(x)s, Y eid %(y)s",
                    {"x": note.eid, "y": aff.eid},
                )
            )
            self.assertTrue(
                cnx.execute(
                    "SET X newinlined Y WHERE X eid %(x)s, Y eid %(y)s",
                    {"x": text.eid, "y": aff.eid},
                )
            )
            self.assertTrue(
                cnx.execute(
                    "SET X newinlined Y WHERE X eid %(x)s, Y eid %(y)s",
                    {"x": note.eid, "y": aff.eid},
                )
            )
            # XXX remove specializes by ourselves, else tearDown fails when removing
            # Para because of Note inheritance. This could be fixed by putting the
            # MemSchemaCWETypeDel(session, name) operation in the
            # after_delete_entity(CWEType) hook, since in that case the MemSchemaSpecializesDel
            # operation would be removed before, but I'm not sure this is a desired behaviour.
            #
            # also we need more tests about introducing/removing base classes or
            # specialization relationship...
            cnx.execute('DELETE X specializes Y WHERE Y name "Para"')
            cnx.commit()
            self.assertEqual(
                sorted(et.type for et in self.schema["Para"].specialized_by()), []
            )
            self.assertEqual(self.schema["Note"].specializes(), None)
            self.assertEqual(self.schema["Text"].specializes(), None)

    def test_add_symmetric_relation_type(self):
        with self.mh() as (cnx, mh):
            self.assertFalse(self.table_sql(mh, "same_as_relation"))
            mh.cmd_add_relation_type("same_as")
            self.assertTrue(self.table_sql(mh, "same_as_relation"))

    def test_change_attribute_type(self):
        with self.mh() as (cnx, mh):
            mh.cmd_create_entity("Societe", tel=1)
            mh.commit()
            mh.change_attribute_type("Societe", "tel", "Float")
            self.assertNotIn(
                ("Societe", "Int"), self.schema["tel"].relation_definitions
            )
            self.assertIn(("Societe", "Float"), self.schema["tel"].relation_definitions)
            self.assertEqual(
                self.schema["tel"].relation_definitions[("Societe", "Float")].object,
                "Float",
            )
            tel = mh.rqlexec("Any T WHERE X tel T")[0][0]
            self.assertEqual(tel, 1.0)
            self.assertIsInstance(tel, float)

    def test_drop_required_inlined_relation(self):
        with self.mh() as (cnx, mh):
            bob = mh.cmd_create_entity("Personne", nom="bob")
            mh.cmd_create_entity("Note", ecrit_par=bob)
            mh.commit()
            rdef = mh.fs_schema.relation_schema_for("ecrit_par").relation_definitions[
                ("Note", "Personne")
            ]
            with tempattr(rdef, "cardinality", "1*"):
                mh.sync_schema_props_perms("ecrit_par", syncperms=False)
            mh.cmd_drop_relation_type("ecrit_par")
            self.assertNotIn(
                f"{SQL_PREFIX}ecrit_par",
                self.table_schema(mh, f"{SQL_PREFIX}Personne"),
            )

    def test_drop_inlined_rdef_delete_data(self):
        with self.mh() as (cnx, mh):
            mh.cmd_create_entity("Note", ecrit_par=cnx.user.eid)
            mh.commit()
            mh.drop_relation_definition("Note", "ecrit_par", "CWUser")
            self.assertFalse(
                mh.sqlexec("SELECT * FROM cw_Note WHERE cw_ecrit_par IS NOT NULL")
            )

    def test_storage_changed(self):
        with self.mh() as (cnx, mh):
            john = mh.cmd_create_entity(
                "Personne", nom="john", photo=Binary(b"something")
            )
            bill = mh.cmd_create_entity("Personne", nom="bill")
            mh.commit()
            with TemporaryDirectory() as tempdir:
                bfs_storage = storages.BytesFileSystemStorage(tempdir)
                storages.set_attribute_storage(
                    self.repo, "Personne", "photo", bfs_storage
                )
                mh.cmd_storage_changed("Personne", "photo")
                bob = mh.cmd_create_entity("Personne", nom="bob")
                bffss_dir_content = os.listdir(tempdir)
                self.assertEqual(len(bffss_dir_content), 1)
                john.cw_clear_all_caches()
                self.assertEqual(
                    john.photo.getvalue(),
                    osp.join(tempdir, bffss_dir_content[0]).encode("utf8"),
                )
                bob.cw_clear_all_caches()
                self.assertIsNone(bob.photo)
                bill.cw_clear_all_caches()
                self.assertIsNone(bill.photo)
                storages.unset_attribute_storage(self.repo, "Personne", "photo")


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
#
# Cubicweb documentation build configuration file, created by
# sphinx-quickstart on Fri Oct 31 09:10:36 2008.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# The contents of this file are pickled, so don't put values in the namespace
# that aren't pickleable (module imports are okay, they're removed automatically).
#
# All configuration values have a default value; values that are commented out
# serve to show the default value.

from os import path as osp

from datetime import date

path = __file__
path = osp.dirname(path)  # ./doc
path = osp.dirname(path)  # ./
path = osp.join(path, "cubicweb", "__pkginfo__.py")  # ./__pkginfo__.py
cw = {}
with open(path) as f:
    exec(f.read(), cw)

# If your extensions are in another directory, add it here. If the directory
# is relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
# sys.path.append(os.path.abspath('some/directory'))

# General configuration
# ---------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
]

autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
# templates_path = []

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General substitutions.
project = "CubicWeb"
copyright = f"2001-{date.today().year}, Logilab"

# The default replacements for |version| and |release|, also used in various
# other places throughout the built documents.
#
# The short X.Y version.
version = ".".join(str(n) for n in cw["numversion"][:2])
# The full version, including alpha/beta/rc tags.
release = cw["version"]

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%B %d, %Y"

# List of documents that shouldn't be included in the build.
unused_docs = []

# A list of glob-style patterns that should be excluded when looking
# for source files. [1] They are matched against the source file names
# relative to the source directory, using slashes as directory
# separators on all platforms.
exclude_patterns = ["book/_maybe_to_integrate"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# Options for HTML output
# -----------------------

# The style sheet to use for HTML and HTML Help pages. A file of that name
# must exist either in Sphinx' static/ path, or in one of the custom paths
# given in html_static_path.
# html_style =

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = f"{project} {release}"

html_theme = "sphinx_book_theme"

html_theme_options = {
    "logo_only": True,
    "repository_url": "https://forge.extranet.logilab.fr/cubicweb/cubicweb",
    "repository_provider": "gitlab",
    "use_repository_button": True,
}

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (within the static path) to place at the top of
# the sidebar.
html_logo = "_static/logo-cubicweb.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_use_modindex = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, the reST sources are included in the HTML build as _sources/<name>.
# html_copy_source = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
html_file_suffix = ".html"

# Output file base name for HTML help builder.
htmlhelp_basename = "Cubicwebdoc"


# Options for LaTeX output
# ------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, document class [howto/manual]).
latex_documents = [
    ("index", "Cubicweb.tex", "Cubicweb Documentation", "Logilab", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True

# aafig_format = dict(latex='pdf', html='svg', text=None)

rst_epilog = """
.. |cubicweb| replace:: *CubicWeb*
.. |yams| replace:: *Yams*
.. |rql| replace:: *RQL*
"""


def setup(app):
    app.add_object_type(
        "confval",
        "confval",
        objname="configuration value",
        indextemplate="pair: %s; configuration value",
    )



# copyright 2003-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.utils"""

import base64
import datetime
import decimal
import doctest
import re
from unittest import TestCase

from cubicweb import Binary, Unauthorized
from cubicweb.entity import Entity
from cubicweb.utils import make_uid, UStringIO, RepeatList

try:
    from cubicweb.utils import CubicWebJsonEncoder, json
except ImportError:
    json = None


class MakeUidTC(TestCase):
    def test_1(self):
        self.assertNotEqual(make_uid("xyz"), make_uid("abcd"))
        self.assertNotEqual(make_uid("xyz"), make_uid("xyz"))

    def test_2(self):
        d = set()
        while len(d) < 10000:
            uid = make_uid("xyz")
            if uid in d:
                self.fail(len(d))
            if re.match(r"\d", uid):
                self.fail(
                    "make_uid must not return something begining with "
                    "some numeric character, got %s" % uid
                )
            d.add(uid)


class UStringIOTC(TestCase):
    def test_boolean_value(self):
        self.assertTrue(UStringIO())


class RepeatListTC(TestCase):
    def test_base(self):
        r_list = RepeatList(3, (1, 3))
        self.assertEqual(r_list[0], (1, 3))
        self.assertEqual(r_list[2], (1, 3))
        self.assertEqual(r_list[-1], (1, 3))
        self.assertEqual(len(r_list), 3)
        # XXX
        self.assertEqual(r_list[4], (1, 3))

        self.assertFalse(RepeatList(0, None))

    def test_slice(self):
        r_list = RepeatList(3, (1, 3))
        self.assertEqual(r_list[0:1], [(1, 3)])
        self.assertEqual(r_list[0:4], [(1, 3)] * 3)
        self.assertEqual(r_list[:], [(1, 3)] * 3)

    def test_iter(self):
        self.assertEqual(list(RepeatList(3, (1, 3))), [(1, 3)] * 3)

    def test_add(self):
        r_list = RepeatList(3, (1, 3))
        self.assertEqual(r_list + [(1, 4)], [(1, 3)] * 3 + [(1, 4)])
        self.assertEqual([(1, 4)] + r_list, [(1, 4)] + [(1, 3)] * 3)
        self.assertEqual(r_list + RepeatList(2, (2, 3)), [(1, 3)] * 3 + [(2, 3)] * 2)

        x = r_list + RepeatList(2, (1, 3))
        self.assertIsInstance(x, RepeatList)
        self.assertEqual(len(x), 5)
        self.assertEqual(x[0], (1, 3))

        x = r_list + [(1, 3)] * 2
        self.assertEqual(x, [(1, 3)] * 5)

    def test_eq(self):
        self.assertEqual(RepeatList(3, (1, 3)), [(1, 3)] * 3)

    def test_pop(self):
        r_list = RepeatList(3, (1, 3))
        r_list.pop(2)
        self.assertEqual(r_list, [(1, 3)] * 2)


class JSONEncoderTC(TestCase):
    def setUp(self):
        if json is None:
            self.skipTest("json not available")

    def encode(self, value):
        return json.dumps(value, cls=CubicWebJsonEncoder)

    def test_encoding_dates(self):
        self.assertEqual(
            self.encode(datetime.datetime(2009, 9, 9, 20, 30)), '"2009/09/09 20:30:00"'
        )
        self.assertEqual(self.encode(datetime.date(2009, 9, 9)), '"2009/09/09"')
        self.assertEqual(self.encode(datetime.time(20, 30)), '"20:30:00"')

    def test_encoding_decimal(self):
        self.assertEqual(self.encode(decimal.Decimal("1.2")), "1.2")

    def test_encoding_bare_entity(self):
        e = Entity(None)
        e.cw_attr_cache["pouet"] = "hop"
        e.eid = 2
        self.assertEqual(json.loads(self.encode(e)), {"pouet": "hop", "eid": 2})

    def test_encoding_entity_in_list(self):
        e = Entity(None)
        e.cw_attr_cache["pouet"] = "hop"
        e.eid = 2
        self.assertEqual(json.loads(self.encode([e])), [{"pouet": "hop", "eid": 2}])

    def test_encoding_binary(self):
        for content in (b"he he", b"h\xe9 hxe9"):
            with self.subTest(content=content):
                encoded = self.encode(Binary(content))
                self.assertEqual(base64.b64decode(encoded), content)

    def test_encoding_unknown_stuff(self):
        self.assertEqual(self.encode(TestCase), "null")


def UnauthorizedTC(TestCase):
    def _test(self, func):
        self.assertEqual(
            func(Unauthorized()), "You are not allowed to perform this operation"
        )
        self.assertEqual(func(Unauthorized("a")), "a")
        self.assertEqual(
            func(Unauthorized("a", "b")),
            "You are not allowed to perform a operation on b",
        )
        self.assertEqual(func(Unauthorized("a", "b", "c")), "a b c")

    def test_str(self):
        self._test(str)


def load_tests(loader, tests, ignore):
    import cubicweb.utils

    tests.addTests(doctest.DocTestSuite(cubicweb.utils))
    return tests


if __name__ == "__main__":
    import unittest

    unittest.main()



from yams import ValidationError
from cubicweb.pyramid.test import PyramidCWTest


def put_in_uncommitable_state(request):
    try:
        request.cw_cnx.execute('SET U login NULL WHERE U login "anon"')
    except ValidationError:
        pass
    request.response.body = b"OK"
    return request.response


class CoreTest(PyramidCWTest):
    anonymous_allowed = True

    def includeme(self, config):
        config.add_route("uncommitable", "/uncommitable")
        config.add_view(put_in_uncommitable_state, route_name="uncommitable")

    def test_uncommitable_cnx(self):
        res = self.webapp.get("/uncommitable")
        self.assertEqual(res.text, "OK")
        self.assertEqual(res.status_int, 200)


if __name__ == "__main__":
    from unittest import main

    main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unittests for cw.devtools.testlib module"""
from logilab.common.registry import yes
from logilab.common.testlib import unittest_main

from cubicweb.devtools.testlib import CubicWebTC


class RepoInstancesConsistencyTC(CubicWebTC):
    test_db_id = "RepoInstancesConsistencyTC"

    def pre_setup_database(self, cnx, config):
        self.assertIs(cnx.repo, config.repository())

    def test_pre_setup(self):
        pass


class CWUtilitiesTC(CubicWebTC):
    def test_temporary_permissions_eschema(self):
        eschema = self.schema["CWUser"]
        with self.temporary_permissions(CWUser={"read": ()}):
            self.assertEqual(eschema.permissions["read"], ())
            self.assertTrue(eschema.permissions["add"])
        self.assertTrue(eschema.permissions["read"], ())

    def test_temporary_permissions_rdef(self):
        rdef = self.schema["CWUser"].relation_definition("in_group")
        with self.temporary_permissions((rdef, {"read": ()})):
            self.assertEqual(rdef.permissions["read"], ())
            self.assertTrue(rdef.permissions["add"])
        self.assertTrue(rdef.permissions["read"], ())

    def test_temporary_permissions_rdef_with_exception(self):
        rdef = self.schema["CWUser"].relation_definition("in_group")
        try:
            with self.temporary_permissions((rdef, {"read": ()})):
                self.assertEqual(rdef.permissions["read"], ())
                self.assertTrue(rdef.permissions["add"])
                raise ValueError("goto")
        except ValueError:
            self.assertTrue(rdef.permissions["read"], ())
        else:
            self.fail("exception was caught unexpectedly")

    def test_temporary_appobjects_registered(self):
        class AnAppobject:
            __registries__ = ("hip",)
            __regid__ = "hop"
            __select__ = yes()
            registered = None

            @classmethod
            def __registered__(cls, reg):
                cls.registered = reg

        with self.temporary_appobjects(AnAppobject):
            self.assertEqual(self.vreg["hip"], AnAppobject.registered)
            self.assertIn(AnAppobject, self.vreg["hip"]["hop"])
        self.assertNotIn(AnAppobject, self.vreg["hip"]["hop"])

    def test_login(self):
        """Calling login should not break hook control"""
        with self.admin_access.repo_cnx() as cnx:
            self.hook_executed = False
            self.create_user(cnx, "babar")
            cnx.commit()

        from cubicweb.server import hook
        from cubicweb.predicates import is_instance

        class MyHook(hook.Hook):
            __regid__ = "whatever"
            __select__ = hook.Hook.__select__ & is_instance("CWProperty")
            category = "test-hook"
            events = ("after_add_entity",)
            test = self

            def __call__(self):
                self.test.hook_executed = True

        with self.new_access("babar").repo_cnx() as cnx:
            with self.temporary_appobjects(MyHook):
                with cnx.allow_all_hooks_but("test-hook"):
                    cnx.create_entity("CWProperty", pkey="ui.language", value="en")
                    cnx.commit()
                    self.assertFalse(self.hook_executed)


class RepoAccessTC(CubicWebTC):
    def test_repo_connection(self):
        acc = self.new_access("admin")
        with acc.repo_cnx() as cnx:
            rset = cnx.execute("Any X WHERE X is CWUser")
            self.assertTrue(rset)

    def test_client_connection(self):
        acc = self.new_access("admin")
        with acc.client_cnx() as cnx:
            rset = cnx.execute("Any X WHERE X is CWUser")
            self.assertTrue(rset)

    def test_admin_access(self):
        with self.admin_access.client_cnx() as cnx:
            self.assertEqual("admin", cnx.user.login)


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""this module contains base classes and utilities for integration with running
http server
"""

import http.client
import random
import socket
from urllib.parse import urlparse

from webtest.http import StopableWSGIServer

from cubicweb.pyramid.test import _BasePyramidCWTest


def get_available_port(ports_scan):
    """return the first available port from the given ports range

    Try to connect port by looking for refused connection (111) or transport
    endpoint already connected (106) errors

    Raise a RuntimeError if no port can be found

    :type ports_range: list
    :param ports_range: range of ports to test
    :rtype: int

    .. see:: :func:`test.test_support.bind_port`
    """
    ports_scan = list(ports_scan)
    random.shuffle(ports_scan)  # lower the chance of race condition
    for port in ports_scan:
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("localhost", port))
        except socket.error as err:
            if err.args[0] in (111, 106):
                return port
        finally:
            s.close()
    raise RuntimeError(
        "get_available_port([ports_range]) cannot find an available port"
    )


class CubicWebServerTC(_BasePyramidCWTest):
    """Base class for running a test web server."""

    ports_range = range(7000, 8000)

    def start_server(self):
        app = self._generate_pyramid_config().make_wsgi_app()

        self.httpd = StopableWSGIServer.create(app, port=self.config["port"])
        self.httpd.wait()

        parseurl = urlparse(self.config["base-url"])
        assert parseurl.port == self.config["port"], (
            self.config["base-url"],
            self.config["port"],
        )

        self._web_test_cnx = http.client.HTTPConnection(
            parseurl.hostname, parseurl.port
        )
        self._ident_cookie = None

    def stop_server(self, timeout=15):
        if self._web_test_cnx is None:
            self.web_logout()
            self._web_test_cnx.close()
        self.httpd.shutdown()

    def web_login(self, user=None, passwd=None):
        """Log the current http session for the provided credential

        If no user is provided, admin connection are used.
        """
        if user is None:
            user = self.admlogin
            passwd = self.admpassword
        if passwd is None:
            passwd = user
        response = self.web_get("login")
        cookie = {
            x.split("=")[0]: x.split("=")[1]
            for x in "; ".join(response.headers.get_all("Set-Cookie")).split("; ")
            if "=" in x
        }
        self._ident_cookie = response.getheader("Set-Cookie")
        assert self._ident_cookie, self._ident_cookie
        response = self.web_request(
            "login",
            body=f"?__login={user}&__password={passwd}&csrf_token={cookie['csrf_token']}",
            method="POST",
        )
        assert response.status == http.client.OK, response.status
        return True

    def web_logout(self, user="admin", pwd=None):
        """Log out current http user"""
        if self._ident_cookie is not None:
            self.web_get("logout")
        self._ident_cookie = None

    def web_request(self, path="", method="GET", body=None, headers=None):
        """Return an http.client.HTTPResponse object for the specified path

        Use available credential if available.
        """
        if headers is None:
            headers = {}
        if self._ident_cookie is not None:
            assert "Cookie" not in headers
            headers["Cookie"] = self._ident_cookie
        self._web_test_cnx.request(method, "/" + path, headers=headers, body=body)
        response = self._web_test_cnx.getresponse()
        response.body = response.read()  # to chain request
        response.read = lambda: response.body
        return response

    def web_get(self, path="", body=None, headers=None):
        return self.web_request(path=path, body=body, headers=headers)

    def setUp(self):
        super().setUp()
        port = get_available_port(self.ports_range)
        self.config.global_set_option("port", port)  # force rewrite here
        self.config.global_set_option("base-url", "http://127.0.0.1:%d/" % port)
        # call load_configuration again to let the config reset its datadir_url
        self.config.load_configuration()
        self.start_server()

    def tearDown(self):
        self.stop_server()
        super().tearDown()



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""hooks to ensure use_email / primary_email relations consistency"""


from cubicweb.server import hook


class SetUseEmailRelationOp(hook.Operation):
    """delay this operation to commit to avoid conflict with a late rql query
    already setting the relation
    """

    rtype = "use_email"
    entity = email = None  # make pylint happy

    def condition(self):
        """check entity has use_email set for the email address"""
        return not any(e for e in self.entity.use_email if self.email.eid == e.eid)

    def precommit_event(self):
        if self.cnx.deleted_in_transaction(self.entity.eid):
            return
        if self.cnx.deleted_in_transaction(self.email.eid):
            return
        if self.condition():
            self.cnx.execute(
                "SET X %s Y WHERE X eid %%(x)s, Y eid %%(y)s" % self.rtype,
                {"x": self.entity.eid, "y": self.email.eid},
            )


class SetPrimaryEmailRelationOp(SetUseEmailRelationOp):
    rtype = "primary_email"

    def condition(self):
        """check entity has no primary_email set"""
        return not self.entity.primary_email


class SetPrimaryEmailHook(hook.Hook):
    """notify when a bug or story or version has its state modified"""

    __regid__ = "setprimaryemail"
    __select__ = hook.Hook.__select__ & hook.match_rtype("use_email")
    category = "email"
    events = ("after_add_relation",)

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        if "primary_email" in list(entity.e_schema.subject_relations.values()):
            SetPrimaryEmailRelationOp(
                self._cw, entity=entity, email=self._cw.entity_from_eid(self.eidto)
            )


class SetUseEmailHook(hook.Hook):
    """notify when a bug or story or version has its state modified"""

    __regid__ = "setprimaryemail"
    __select__ = hook.Hook.__select__ & hook.match_rtype("primary_email")
    category = "email"
    events = ("after_add_relation",)

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        if "use_email" in list(entity.e_schema.subject_relations.values()):
            SetUseEmailRelationOp(
                self._cw, entity=entity, email=self._cw.entity_from_eid(self.eidto)
            )



# pylint: disable=W0622,C0103
# copyright 2003-2020 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb global packaging information for the cubicweb knowledge management
software
"""

modname = distname = "cubicweb"

numversion = (4, 6, 1)
version = ".".join(str(num) for num in numversion)

description = "a repository of entities / relations for knowledge management"
author = "Logilab"
author_email = "contact@logilab.fr"
web = "https://www.cubicweb.org"
license = "LGPL"

classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: JavaScript",
]

# data files that shall be copied into the main package directory
package_data = {
    "cubicweb.web.views": ["*.pt"],
    "cubicweb.pyramid": ["development.ini.tmpl"],
}



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Functions to add additional annotations on a rql syntax tree to ease later
code generation.
"""

from rql import BadRQLQuery
from rql.nodes import Relation, VariableRef, Constant, Variable, Or
from rql.utils import common_parent


class CantSelectPrincipal(Exception):
    """raised when no 'principal' variable can be found"""


def _select_principal(scope, relations, _sort=lambda x: x):
    """given a list of rqlst relations, select one which will be used to
    represent an invariant variable (e.g. using on extremity of the relation
    instead of the variable's type table
    """
    # _sort argument is there for test
    diffscope_rels = {}
    ored_rels = set()
    diffscope_rels = set()
    for rel, role in _sort(relations):
        # note: only eid and has_text among all final relations may be there
        if rel.r_type in ("eid", "identity"):
            continue
        if rel.optional is not None and len(relations) > 1:
            if role == "subject" and rel.optional == "right":
                continue
            if role == "object" and rel.optional == "left":
                continue
        if rel.ored(traverse_scope=True):
            ored_rels.add(rel)
        elif rel.scope is scope:
            return rel
        elif not rel.neged(traverse_scope=True):
            diffscope_rels.add(rel)
    if len(ored_rels) > 1:
        ored_rels_copy = tuple(ored_rels)
        for rel1 in ored_rels_copy:
            for rel2 in ored_rels_copy:
                if rel1 is rel2:
                    continue
                if isinstance(common_parent(rel1, rel2), Or):
                    ored_rels.discard(rel1)
                    ored_rels.discard(rel2)
    for rel in _sort(ored_rels):
        if rel.scope is scope:
            return rel
        diffscope_rels.add(rel)
    # if DISTINCT query, can use variable from a different scope as principal
    # since introduced duplicates will be removed
    if scope.stmt.distinct and diffscope_rels:
        return next(iter(_sort(diffscope_rels)))
    # XXX could use a relation from a different scope if it can't generate
    # duplicates, so we should have to check cardinality
    raise CantSelectPrincipal()


def _select_main_var(relations):
    """given a list of rqlst relations, select one which will be used as main
    relation for the rhs variable
    """
    principal = None
    others = []
    # sort for test predictability
    for rel in sorted(relations, key=lambda x: (x.children[0].name, x.r_type)):
        # only equality relation with a variable as rhs may be principal
        if (
            rel.operator() not in ("=", "IS")
            or not isinstance(rel.children[1].children[0], VariableRef)
            or rel.neged(strict=True)
        ):
            continue
        if rel.optional:
            others.append(rel)
            continue
        if rel.scope is rel.stmt:
            return rel
        principal = rel
    if principal is None:
        if others:
            return others[0]
        raise BadRQLQuery(
            f"unable to find principal in {', '.join(r.as_string() for r in relations)}"
        )
    return principal


def set_qdata(getrschema, union, noinvariant):
    """recursive function to set querier data on variables in the syntax tree"""
    for select in union.children:
        for subquery in select.with_:
            set_qdata(getrschema, subquery.query, noinvariant)
        for var in select.defined_vars.values():
            if var.stinfo["invariant"]:
                if var in noinvariant:
                    var._q_invariant = False
                else:
                    var._q_invariant = True
            else:
                var._q_invariant = False


class RQLAnnotator:
    """Annotate the RQL abstract syntax tree to inform the SQL generation"""

    def __init__(self, schema):
        self.schema = schema
        self.nfdomain = frozenset(
            eschema.type for eschema in schema.entities() if not eschema.final
        )

    def annotate(self, rqlst):
        """add information to the rql syntax tree to help sources to do their
        job (read sql generation)

        a variable is tagged as invariant if:
        * it is a non final variable
        * it is not used as lhs in any final or inlined relation
        * there is no type restriction on this variable (either explicit in the
          syntax tree or because a solution for this variable has been removed
          due to security filtering)
        """
        # assert rqlst.TYPE == 'select', rqlst
        rqlst.has_text_query = self._annotate_union(rqlst)

    def _annotate_union(self, union):
        has_text_query = False
        for select in union.children:
            if self._annotate_select(select):
                has_text_query = True
        return has_text_query

    def _annotate_select(self, rqlst):
        has_text_query = False
        for subquery in rqlst.with_:
            if self._annotate_union(subquery.query):
                has_text_query = True
        getrschema = self.schema.relation_schema_for
        for var in rqlst.defined_vars.values():
            stinfo = var.stinfo
            if stinfo.get("ftirels"):
                has_text_query = True
            if stinfo["attrvar"]:
                stinfo["invariant"] = False
                stinfo["principal"] = _select_main_var(stinfo["rhsrelations"])
                continue
            if stinfo["typerel"] is None:
                # those particular queries should be executed using the system
                # entities table unless there is some type restriction
                if not stinfo["relations"]:
                    # Any X, Any MAX(X)...
                    stinfo["invariant"] = True
                    stinfo["principal"] = None
                    continue
                if any(
                    rel
                    for rel in stinfo["relations"]
                    if rel.r_type == "eid" and rel.operator() != "="
                ) and not any(
                    r
                    for r in var.stinfo["relations"] - var.stinfo["rhsrelations"]
                    if r.r_type != "eid"
                    and (getrschema(r.r_type).inlined or getrschema(r.r_type).final)
                ):
                    # Any X WHERE X eid > 2
                    stinfo["invariant"] = True
                    stinfo["principal"] = None
                    continue
            if stinfo["selected"] and var.valuable_references() == 1 + bool(
                stinfo["constnode"]
            ):
                # "Any X", "Any X, Y WHERE X attr Y"
                stinfo["invariant"] = False
                continue
            joins = set()
            invariant = False
            for ref in var.references():
                rel = ref.relation()
                if rel is None or rel.is_types_restriction():
                    continue
                lhs, rhs = rel.get_parts()
                onlhs = ref is lhs
                role = "subject" if onlhs else "object"
                if rel.r_type == "eid":
                    if not (onlhs and len(stinfo["relations"]) > 1):
                        break
                    if not stinfo["constnode"]:
                        joins.add((rel, role))
                    continue
                elif rel.r_type == "identity":
                    # identity can't be used as principal, so check other relation are used
                    # XXX explain rhs.operator == '='
                    if rhs.operator != "=" or len(stinfo["relations"]) <= 1:
                        break
                    joins.add((rel, role))
                    continue
                rschema = getrschema(rel.r_type)
                if rel.optional:
                    if rel in stinfo.get("optrelations", ()):
                        # optional variable can't be invariant if this is the lhs
                        # variable of an inlined relation
                        if rel not in stinfo["rhsrelations"] and rschema.inlined:
                            break
                    # variable used as main variable of an optional relation can't
                    # be invariant, unless we can use some other relation as
                    # reference for the outer join
                    elif not stinfo["constnode"]:
                        break
                    elif len(stinfo["relations"]) == 2:
                        if onlhs:
                            ostinfo = rhs.children[0].variable.stinfo
                        else:
                            ostinfo = lhs.variable.stinfo
                        if not (
                            ostinfo.get("optcomparisons")
                            or any(
                                orel
                                for orel in ostinfo["relations"]
                                if orel.optional and orel is not rel
                            )
                        ):
                            break
                if rschema.final or (onlhs and rschema.inlined):
                    if rschema.type != "has_text":
                        # need join anyway if the variable appears in a final or
                        # inlined relation
                        break
                    joins.add((rel, role))
                    continue
                if not stinfo["constnode"]:
                    if rschema.inlined and rel.neged(strict=True):
                        # if relation is inlined, can't be invariant if that
                        # variable is used anywhere else.
                        # see 'Any P WHERE NOT N ecrit_par P, N eid 512':
                        # sql for 'NOT N ecrit_par P' is 'N.ecrit_par is NULL' so P
                        # can use N.ecrit_par as principal
                        if stinfo["selected"] or len(stinfo["relations"]) > 1:
                            break
                joins.add((rel, role))
            else:
                # if there is at least one ambigous relation and no other to
                # restrict types, can't be invariant since we need to filter out
                # other types
                if not self.is_ambiguous(var):
                    invariant = True
            stinfo["invariant"] = invariant
            if invariant and joins:
                # remember rqlst/solutions analyze information
                # we have to select a kindof "main" relation which will "extrajoins"
                # the other
                # priority should be given to relation which are not in inner queries
                # (eg exists)
                try:
                    stinfo["principal"] = principal = _select_principal(
                        var.scope, joins
                    )
                    if getrschema(principal.r_type).inlined:
                        # the scope of the lhs variable must be equal or outer to the
                        # rhs variable's scope (since it's retrieved from lhs's table)
                        sstinfo = principal.children[0].variable.stinfo
                        sstinfo["scope"] = common_parent(
                            sstinfo["scope"], stinfo["scope"]
                        ).scope
                except CantSelectPrincipal:
                    stinfo["invariant"] = False
        # see unittest_rqlannotation. test_has_text_security_cache_bug
        # XXX probably more to do, but yet that work without more...
        for col_alias in rqlst.aliases.values():
            if col_alias.stinfo.get("ftirels"):
                has_text_query = True
        return has_text_query

    def is_ambiguous(self, var):
        # ignore has_text relation when we know it will be used as principal.
        # This is expected by the rql2sql generator which will use the `entities`
        # table to filter out by type if necessary, This optimisation is very
        # interesting in multi-sources cases, as it may avoid a costly query
        # on sources to get all entities of a given type to achieve this, while
        # we have all the necessary information.
        root = var.stmt.root  # Union node
        # rel.scope -> Select or Exists node, so add .parent to get Union from
        # Select node
        rels = [rel for rel in var.stinfo["relations"] if rel.scope.parent is root]
        if len(rels) == 1 and rels[0].r_type == "has_text":
            return False
        try:
            data = var.stmt._deamb_data
        except AttributeError:
            data = var.stmt._deamb_data = IsAmbData(self.schema, self.nfdomain)
            data.compute(var.stmt)
        return data.is_ambiguous(var)


class IsAmbData:
    def __init__(self, schema, nfdomain):
        self.schema = schema
        # shortcuts
        self.relation_schema_for = schema.relation_schema_for
        self.entity_schema_for = schema.entity_schema_for
        # domain for non final variables
        self.nfdomain = nfdomain
        # {var: possible solutions set}
        self.varsols = {}
        # set of ambiguous variables
        self.ambiguousvars = set()
        # remember if a variable has been deambiguified by another to avoid
        # doing the opposite
        self.deambification_map = {}
        # not invariant variables (access to final.inlined relation)
        self.not_invariants = set()

    def is_ambiguous(self, var):
        return var in self.ambiguousvars

    def restrict(self, var, restricted_domain):
        self.varsols[var] &= restricted_domain
        if (
            var in self.ambiguousvars
            and self.varsols[var] == var.stinfo["possibletypes"]
        ):
            self.ambiguousvars.remove(var)

    def compute(self, rqlst):
        # set domains for each variable
        for varname, var in rqlst.defined_vars.items():
            if (
                var.stinfo["uidrel"] is not None
                or self.entity_schema_for(rqlst.solutions[0][varname]).final
            ):
                ptypes = var.stinfo["possibletypes"]
            else:
                ptypes = set(self.nfdomain)
                self.ambiguousvars.add(var)
            self.varsols[var] = ptypes
        if not self.ambiguousvars:
            return
        # apply relation restriction
        self.maydeambrels = maydeambrels = {}
        for rel in rqlst.iget_nodes(Relation):
            if rel.r_type == "eid" or rel.is_types_restriction():
                continue
            lhs, rhs = rel.get_variable_parts()
            if isinstance(lhs, VariableRef) or isinstance(rhs, VariableRef):
                rschema = self.relation_schema_for(rel.r_type)
                if rschema.inlined or rschema.final:
                    self.not_invariants.add(lhs.variable)
                self.set_rel_constraint(lhs, rel, rschema.subjects)
                self.set_rel_constraint(rhs, rel, rschema.objects)
        # try to deambiguify more variables by considering other variables'type
        modified = True
        while modified and self.ambiguousvars:
            modified = False
            for var in self.ambiguousvars.copy():
                try:
                    for rel in var.stinfo["relations"] & maydeambrels[var]:
                        if self.deambiguifying_relation(var, rel):
                            modified = True
                            break
                except KeyError:
                    # no relation to deambiguify
                    continue

    def _debug_print(self):
        print(
            "varsols",
            {x: sorted(str(v) for v in values) for x, values in self.varsols.items()},
        )
        print("ambiguous vars", sorted(self.ambiguousvars))

    def set_rel_constraint(self, term, rel, etypes_func):
        if isinstance(term, VariableRef) and self.is_ambiguous(term.variable):
            var = term.variable
            if (
                len(var.stinfo["relations"]) == 1
                or rel.scope is var.scope
                or rel.r_type == "identity"
            ):
                self.restrict(var, frozenset(etypes_func()))
                try:
                    self.maydeambrels[var].add(rel)
                except KeyError:
                    self.maydeambrels[var] = {rel}

    def deambiguifying_relation(self, var, rel):
        lhs, rhs = rel.get_variable_parts()
        onlhs = var is getattr(lhs, "variable", None)
        other = onlhs and rhs or lhs
        otheretypes = None
        # XXX isinstance(other.variable, Variable) to skip column alias
        if isinstance(other, VariableRef) and isinstance(other.variable, Variable):
            deambiguifier = other.variable
            if var is not self.deambification_map.get(deambiguifier):
                if var.stinfo["typerel"] is None:
                    otheretypes = deambiguifier.stinfo["possibletypes"]
                elif not self.is_ambiguous(deambiguifier):
                    otheretypes = self.varsols[deambiguifier]
                elif deambiguifier in self.not_invariants:
                    # we know variable won't be invariant, try to use
                    # it to deambguify the current variable
                    otheretypes = self.varsols[deambiguifier]
            if deambiguifier.stinfo["typerel"] is None:
                # if deambiguifier has no type restriction using 'is',
                # don't record it
                deambiguifier = None
        elif isinstance(other, Constant) and other.uidtype:
            otheretypes = (other.uidtype,)
            deambiguifier = None
        if otheretypes is not None:
            # to restrict, we must check that for all type in othertypes,
            # possible types on the other end of the relation are matching
            # variable's possible types
            rschema = self.relation_schema_for(rel.r_type)
            if onlhs:
                rtypefunc = rschema.subjects
            else:
                rtypefunc = rschema.objects
            for otheretype in otheretypes:
                reltypes = frozenset(rtypefunc(otheretype))
                if var.stinfo["possibletypes"] != reltypes:
                    return False
            self.restrict(var, var.stinfo["possibletypes"])
            self.deambification_map[var] = deambiguifier
            return True
        return False



# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-forum schema"""

from yams.buildobjs import (
    String,
    RichString,
    EntityType,
    RelationDefinition,
    SubjectRelation,
)


class Forum(EntityType):
    topic = String(maxsize=50, required=True, unique=True)
    description = RichString()


class ForumThread(EntityType):
    __permissions__ = {
        "read": ("managers", "users"),
        "add": ("managers", "users"),
        "update": ("managers", "owners"),
        "delete": ("managers", "owners"),
    }
    title = String(required=True, fulltextindexed=True, maxsize=256)
    content = RichString(required=True, fulltextindexed=True)
    in_forum = SubjectRelation(
        "Forum", cardinality="1*", inlined=True, composite="object"
    )


class interested_in(RelationDefinition):
    subject = "CWUser"
    object = ("ForumThread", "Forum")


class nosy_list(RelationDefinition):
    subject = ("Forum", "ForumThread")
    object = "CWUser"



# copyright 2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Tests for cubicweb.pyramid.auth module."""

from http import cookiejar as http_cookiejar

from pyramid import security
from pyramid.settings import asbool
from pyramid.response import Response
from pyramid.httpexceptions import HTTPSeeOther

from cubicweb.pyramid.test import PyramidCWTest


def login_view(request):
    """
    Dummy testing view that only stimulate login to call out auth policy.
    It doesn't check the password or anything else.
    """
    repo = request.registry["cubicweb.repository"]

    with repo.internal_cnx() as cnx:
        user_eid = cnx.find("CWUser", login="admin")[0][0]

    headers = security.remember(
        request,
        user_eid,
        persistent=asbool(request.params.get("__setauthcookie", False)),
    )

    raise HTTPSeeOther("/", headers=headers)


def logout_view(request):
    headers = security.forget(request)
    raise HTTPSeeOther("/", headers=headers)


def current_user_view(request):
    return Response(request.cw_cnx.user.login)


class PyramidCWAuthTktAuthenticationPolicyTC(PyramidCWTest):
    anonymous_allowed = True

    def includeme(self, config):
        config.include("cubicweb.pyramid.auth")
        config.include("cubicweb.pyramid.session")

        config.add_route("login", "/login")
        config.add_route("logout", "/logout")
        config.add_route("current_user", "/current_user")

        config.add_view(login_view, route_name="login")
        config.add_view(logout_view, route_name="logout")
        config.add_view(current_user_view, route_name="current_user")

    def test_remember(self):
        assert self.webapp.get("/current_user").text == "anon"
        self.webapp.get("/login")
        assert self.webapp.get("/current_user").text == "admin"

    def test_remember_and_forget(self):
        assert self.webapp.get("/current_user").text == "anon"
        self.webapp.get("/login")
        assert self.webapp.get("/current_user").text == "admin"
        self.webapp.get("/logout")
        assert self.webapp.get("/current_user").text == "anon"

    def test_cant_reuse_loggedout_cookie(self):
        assert self.webapp.get("/current_user").text == "anon"
        self.webapp.get("/login")

        # simulate that the cookies have been stolen and that someone attempts to reuse them
        stolen_jar = http_cookiejar.CookieJar()
        for cookie in self.webapp.cookiejar:
            stolen_jar.set_cookie(cookie)

        assert self.webapp.get("/current_user").text == "admin"
        self.webapp.get("/logout")
        assert self.webapp.get("/current_user").text == "anon"

        self.webapp.cookiejar = stolen_jar

        # check that we can't reuse the cookie
        assert self.webapp.get("/current_user").text == "anon"


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.
"""Instrumentation utilities"""

try:
    import pygraphviz
except ImportError:
    pygraphviz = None

from cubicweb.cwvreg import CWRegistryStore
from cubicweb.devtools.devctl import DevConfiguration


ALL_COLORS = [
    "00FF00",
    "0000FF",
    "FFFF00",
    "FF00FF",
    "00FFFF",
    "000000",
    "800000",
    "008000",
    "000080",
    "808000",
    "800080",
    "008080",
    "808080",
    "C00000",
    "00C000",
    "0000C0",
    "C0C000",
    "C000C0",
    "00C0C0",
    "C0C0C0",
    "400000",
    "004000",
    "000040",
    "404000",
    "400040",
    "004040",
    "404040",
    "200000",
    "002000",
    "000020",
    "202000",
    "200020",
    "002020",
    "202020",
    "600000",
    "006000",
    "000060",
    "606000",
    "600060",
    "006060",
    "606060",
    "A00000",
    "00A000",
    "0000A0",
    "A0A000",
    "A000A0",
    "00A0A0",
    "A0A0A0",
    "E00000",
    "00E000",
    "0000E0",
    "E0E000",
    "E000E0",
    "00E0E0",
    "E0E0E0",
]
_COLORS = {}


def get_color(key):
    try:
        return _COLORS[key]
    except KeyError:
        _COLORS[key] = "#" + ALL_COLORS[len(_COLORS) % len(ALL_COLORS)]
        return _COLORS[key]


def warn(msg, *args):
    print(f"WARNING: {msg % args}")


def info(msg):
    print("INFO: " + msg)


class PropagationAnalyzer:
    """Abstract propagation analyzer, providing utility function to extract
    entities involved in propagation from a schema, as well as propagation
    rules from hooks (provided they use intrumentalized sets, see
    :class:`CubeTracerSet`).

    Concrete classes should at least define `prop_rel` class attribute and
    implements the `is_root` method.

    See `localperms` or `nosylist` cubes for example usage (`ccplugin` module).
    """

    prop_rel = None  # name of the propagation relation

    def init(self, cube):
        """Initialize analyze for the given cube, returning the (already loaded)
        vregistry and a set of entities which we're interested in.
        """
        config = DevConfiguration(cube)
        schema = config.load_schema()
        vreg = CWRegistryStore(config)
        vreg.set_schema(schema)  # set_schema triggers objects registrations
        eschemas = {
            eschema for eschema in schema.entities() if self.should_include(eschema)
        }
        return vreg, eschemas

    def is_root(self, eschema):
        """Return `True` if given entity schema is a root of the graph"""
        raise NotImplementedError()

    def should_include(self, eschema):
        """Return `True` if given entity schema should be included by the graph."""

        if self.prop_rel in eschema.subject_relations or self.is_root(eschema):
            return True
        return False

    def prop_edges(self, s_rels, o_rels, eschemas):
        """Return a set of edges where propagation has been detected.

        Each edge is defined by a 4-uple (from node, to node, rtype, package)
        where `rtype` is the relation type bringing from <from node> to <to
        node> and `package` is the cube adding the rule to the propagation
        control set (see see :class:`CubeTracerSet`).
        """
        schema = iter(eschemas).next().schema
        prop_edges = set()
        for rtype in s_rels:
            found = False
            for subj, obj in schema.relation_schema_for(rtype).relation_definitions:
                if subj in eschemas and obj in eschemas:
                    found = True
                    prop_edges.add((subj, obj, rtype, s_rels.value_cube[rtype]))
            if not found:
                warn("no rdef match for %s", rtype)
        for rtype in o_rels:
            found = False
            for subj, obj in schema.relation_schema_for(rtype).relation_definitions:
                if subj in eschemas and obj in eschemas:
                    found = True
                    prop_edges.add((obj, subj, rtype, o_rels.value_cube[rtype]))
            if not found:
                warn("no rdef match for %s", rtype)
        return prop_edges

    def detect_problems(self, eschemas, edges):
        """Given the set of analyzed entity schemas and edges between them,
        return a set of entity schemas where a problem has been detected.
        """
        problematic = set()
        for eschema in eschemas:
            if self.has_problem(eschema, edges):
                problematic.add(eschema)
        not_problematic = set(eschemas).difference(problematic)
        if not_problematic:
            info(
                f"nothing problematic in: {', '.join(e.type for e in not_problematic)}"
            )
        return problematic

    def has_problem(self, eschema, edges):
        """Return `True` if the given schema is considered problematic,
        considering base propagation rules.
        """
        root = self.is_root(eschema)
        has_prop_rel = self.prop_rel in eschema.subject_relations
        # root but no propagation relation
        if root and not has_prop_rel:
            warn("%s is root but miss %s", eschema, self.prop_rel)
            return True
        # propagated but without propagation relation / not propagated but
        # with propagation relation
        if not has_prop_rel and any(edge for edge in edges if edge[1] == eschema):
            warn("%s miss %s but is reached by propagation", eschema, self.prop_rel)
            return True
        elif has_prop_rel and not root:
            rdef = eschema.rdef(self.prop_rel, take_first=True)
            edges = [edge for edge in edges if edge[1] == eschema]
            if not edges:
                warn(
                    "%s has %s but isn't reached by " "propagation",
                    eschema,
                    self.prop_rel,
                )
                return True
            # require_permission relation / propagation rule not added by
            # the same cube
            elif not any(edge for edge in edges if edge[-1] == rdef.package):
                warn(
                    "%s has %s relation / propagation rule"
                    " not added by the same cube (%s / %s)",
                    eschema,
                    self.prop_rel,
                    rdef.package,
                    edges[0][-1],
                )
                return True
        return False

    def init_graph(self, eschemas, edges, problematic):
        """Initialize and return graph, adding given nodes (entity schemas) and
        edges between them.

        Require pygraphviz installed.
        """
        if pygraphviz is None:
            raise RuntimeError("pygraphviz is not installed")
        graph = pygraphviz.AGraph(strict=False, directed=True)
        for eschema in eschemas:
            if eschema in problematic:
                params = {"color": "#ff0000", "fontcolor": "#ff0000"}
            else:
                params = {}  # 'color': get_color(eschema.package)}
            graph.add_node(eschema.type, **params)
        for subj, obj, rtype, package in edges:
            graph.add_edge(str(subj), str(obj), label=rtype, color=get_color(package))
        return graph

    def add_colors_legend(self, graph):
        """Add a legend of used colors to the graph."""
        for package, color in sorted(_COLORS.items()):
            graph.add_node(package, color=color, fontcolor=color, shape="record")


class CubeTracerSet:
    """Dumb set implementation whose purpose is to keep track of which cube is
    being loaded when something is added to the set.

    Results will be found in the `value_cube` attribute dictionary.

    See `localperms` or `nosylist` cubes for example usage (`hooks` module).
    """

    def __init__(self, vreg, wrapped):
        self.vreg = vreg
        self.wrapped = wrapped
        self.value_cube = {}

    def add(self, value):
        self.wrapped.add(value)
        cube = self.vreg.currently_loading_cube
        if value in self.value_cube:
            warn(
                "%s is propagated by cube %s and cube %s",
                value,
                self.value_cube[value],
                cube,
            )
        else:
            self.value_cube[value] = cube

    def __iter__(self):
        return iter(self.wrapped)

    def __ior__(self, other):
        for value in other:
            self.add(value)
        return self

    def __ror__(self, other):
        other |= self.wrapped
        return other



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""SQL utilities functions and classes."""

import os
import re
import subprocess
import sys
from datetime import time, datetime, timedelta
from logging import getLogger
from os.path import abspath

from logilab import database as logilab_database, common as lgc
from logilab.common.date import utctime, utcdatetime
from logilab.common.logging_ext import set_log_methods
from logilab.common.shellutils import ProgressBar, DummyProgressBar
from logilab.database import SQL_CONNECT_HOOKS
from logilab.database.sqlgen import SQLGenerator
from pytz import utc

from cubicweb import Binary, ConfigurationError
from cubicweb.schema import PURE_VIRTUAL_RTYPES
from cubicweb.server.utils import crypt_password
from cubicweb.uilib import remove_html_tags

lgc.USE_MX_DATETIME = False
SQL_PREFIX = "cw_"


def _run_command(cmd, extra_env=None):
    env = os.environ.copy()
    for key, value in (extra_env or {}).items():
        env.setdefault(key, value)
    if isinstance(cmd, str):
        print(cmd)
        return subprocess.call(cmd, shell=True, env=env)
    else:
        print(" ".join(cmd))
        return subprocess.call(cmd, env=env)


def sqlexec(
    sqlstmts, cursor_or_execute, withpb=True, pbtitle="", delimiter=";", cnx=None
):
    """execute sql statements ignoring DROP/ CREATE GROUP or USER statements
    error.

    :sqlstmts_as_string: a string or a list of sql statements.
    :cursor_or_execute: sql cursor or a callback used to execute statements
    :cnx: if given, commit/rollback at each statement.

    :withpb: if True, display a progresse bar
    :pbtitle: a string displayed as the progress bar title (if `withpb=True`)

    :delimiter: a string used to split sqlstmts (if it is a string)

    Return the failed statements (same type as sqlstmts)
    """
    if hasattr(cursor_or_execute, "execute"):
        execute = cursor_or_execute.execute
    else:
        execute = cursor_or_execute
    sqlstmts_as_string = False
    if isinstance(sqlstmts, str):
        sqlstmts_as_string = True
        sqlstmts = sqlstmts.split(delimiter)
    if withpb:
        if sys.stdout.isatty():
            pb = ProgressBar(len(sqlstmts), title=pbtitle)
        else:
            pb = DummyProgressBar()
    failed = []
    for sql in sqlstmts:
        sql = sql.strip()
        if withpb:
            pb.update()
        if not sql:
            continue
        try:
            # some dbapi modules doesn't accept unicode for sql string
            execute(str(sql))
        except Exception as ex:
            print(ex, file=sys.stderr)
            if cnx:
                cnx.rollback()
            failed.append(sql)
        else:
            if cnx:
                cnx.commit()
    if withpb:
        print()
    if sqlstmts_as_string:
        failed = delimiter.join(failed)
    return failed


def sqlgrants(
    schema,
    driver,
    user,
    text_index=True,
    set_owner=True,
    skip_relations=(),
    skip_entities=(),
):
    """Return a list of SQL statements to give all access privileges to the given user on the
    database.
    """
    from cubicweb.server.schema2sql import grant_schema
    from cubicweb.server.sources import native

    stmts = list(native.grant_schema(user, set_owner))
    if text_index:
        dbhelper = logilab_database.get_db_helper(driver)
        # XXX should return a list of sql statements rather than ';' joined statements
        stmts += dbhelper.sql_grant_user_on_fti(user).split(";")
    stmts += grant_schema(
        schema, user, set_owner, skip_entities=skip_entities, prefix=SQL_PREFIX
    )
    return stmts


def sqlschema(
    schema,
    driver,
    text_index=True,
    user=None,
    set_owner=False,
    skip_relations=PURE_VIRTUAL_RTYPES,
    skip_entities=(),
):
    """
    Return the database SQL schema as a list of SQL statements, according to the given parameters.
    """
    from cubicweb.server.schema2sql import schema2sql
    from cubicweb.server.sources import native

    if set_owner:
        assert user, "user is argument required when set_owner is true"
    stmts = list(native.sql_schema(driver))
    dbhelper = logilab_database.get_db_helper(driver)
    if text_index:
        stmts += dbhelper.sql_init_fti().split(";")  # XXX
    stmts += schema2sql(
        dbhelper,
        schema,
        prefix=SQL_PREFIX,
        skip_entities=skip_entities,
        skip_relations=skip_relations,
    )
    if dbhelper.users_support and user:
        stmts += sqlgrants(
            schema, driver, user, text_index, set_owner, skip_relations, skip_entities
        )
    return stmts


_SQL_DROP_ALL_USER_TABLES_FILTER_FUNCTION = re.compile("^(?!(sql|pg)_)").match


def sql_drop_all_user_tables(driver_or_helper, sqlcursor):
    """Return ths sql to drop all tables found in the database system."""
    if not getattr(driver_or_helper, "list_tables", None):
        dbhelper = logilab_database.get_db_helper(driver_or_helper)
    else:
        dbhelper = driver_or_helper

    stmts = [dbhelper.sql_drop_sequence("entities_id_seq")]
    # for mssql, we need to drop views before tables
    if hasattr(dbhelper, "list_views"):
        stmts += [
            f"DROP VIEW {name};"
            for name in filter(
                _SQL_DROP_ALL_USER_TABLES_FILTER_FUNCTION,
                dbhelper.list_views(sqlcursor),
            )
        ]
    stmts += [
        f"DROP TABLE {name};"
        for name in filter(
            _SQL_DROP_ALL_USER_TABLES_FILTER_FUNCTION, dbhelper.list_tables(sqlcursor)
        )
    ]
    return stmts


class ConnectionWrapper:
    """Wrap a connection to the system source's database, attempting to handle
    automatic reconnection.
    """

    # since 3.19, we only have to manage the system source connection
    def __init__(self, system_source):
        # dictionary of (source, connection), indexed by sources'uri
        self._source = system_source
        self.cnx = system_source.get_connection()
        self.cu = self.cnx.cursor()

    def commit(self):
        """commit the current transaction for this user"""
        # let exception propagates
        self.cnx.commit()

    def rollback(self):
        """rollback the current transaction for this user"""
        # catch exceptions, rollback other sources anyway
        try:
            self.cnx.rollback()
        except Exception:
            self._source.critical("rollback error", exc_info=sys.exc_info())
            # error on rollback, the connection is much probably in a really
            # bad state. Replace it by a new one.
            self.reconnect()

    def close(self, i_know_what_i_do=False):
        """close all connections in the set"""
        if i_know_what_i_do is not True:  # unexpected closing safety belt
            raise RuntimeError("connections set shouldn't be closed")
        try:
            self.cu.close()
            self.cu = None
        except Exception:
            pass
        try:
            self.cnx.close()
            self.cnx = None
        except Exception:
            pass

    # internals ###############################################################

    def cnxset_freed(self):
        """connections set is being freed from a session"""
        pass  # no nothing by default

    def reconnect(self):
        """reopen a connection for this source or all sources if none specified"""
        try:
            # properly close existing connection if any
            self.cnx.close()
        except Exception:
            pass
        self._source.info("trying to reconnect")
        self.cnx = self._source.get_connection()
        self.cu = self.cnx.cursor()


class SqliteConnectionWrapper(ConnectionWrapper):
    """Sqlite specific connection wrapper: close the connection each time it's
    freed (and reopen it later when needed)
    """

    def __init__(self, system_source):
        # don't call parent's __init__, we don't want to initiate the connection
        self._source = system_source

    _cnx = None

    def cnxset_freed(self):
        self.cu.close()
        self.cnx.close()
        self.cnx = self.cu = None

    @property
    def cnx(self):
        if self._cnx is None:
            self._cnx = self._source.get_connection()
            self._cu = self._cnx.cursor()
        return self._cnx

    @cnx.setter
    def cnx(self, value):
        self._cnx = value

    @property
    def cu(self):
        if self._cnx is None:
            self._cnx = self._source.get_connection()
            self._cu = self._cnx.cursor()
        return self._cu

    @cu.setter
    def cu(self, value):
        self._cu = value


class SQLAdapterMixIn:
    """Mixin for SQL data sources, getting a connection from a configuration
    dictionary and handling connection locking
    """

    def __init__(self, source_config, repairing=False):
        try:
            self.dbdriver = source_config["db-driver"].lower()
            dbname = source_config["db-name"]
        except KeyError as e:
            raise ConfigurationError(
                "missing some expected entries in sources file (do you have "
                "a db-driver and a db-name keys?), error: %s" % e
            )

        dbhost = source_config.get("db-host")
        port = source_config.get("db-port")
        dbport = port and int(port) or None
        dbuser = source_config.get("db-user")
        dbpassword = source_config.get("db-password")
        dbencoding = source_config.get("db-encoding", "UTF-8")
        dbextraargs = source_config.get("db-extra-arguments")
        dbnamespace = source_config.get("db-namespace")

        self.dbhelper = logilab_database.get_db_helper(self.dbdriver)
        self.dbhelper.record_connection_info(
            dbname,
            dbhost,
            dbport,
            dbuser,
            dbpassword,
            dbextraargs,
            dbencoding,
            dbnamespace,
        )
        self.sqlgen = SQLGenerator()

        # copy back some commonly accessed attributes
        dbapi_module = self.dbhelper.dbapi_module
        self.OperationalError = dbapi_module.OperationalError
        self.InterfaceError = dbapi_module.InterfaceError
        self.DbapiError = dbapi_module.Error
        self._binary = self.dbhelper.binary_value
        self._process_value = dbapi_module.process_value
        self._dbencoding = dbencoding

        if self.dbdriver == "sqlite":
            self.cnx_wrap = SqliteConnectionWrapper
            self.dbhelper.dbname = abspath(self.dbhelper.dbname)
        else:
            self.cnx_wrap = ConnectionWrapper

        if not repairing:
            db_statement_timeout = source_config.get("db-statement-timeout", 0)
            # If we have db-statement-timeout= in sources, then the key exists
            # in source_config, but its value is None, so we need to check it.
            statement_timeout = int(db_statement_timeout if db_statement_timeout else 0)

            if statement_timeout > 0:

                def set_postgres_timeout(cnx):
                    cnx.cursor().execute(
                        "SET statement_timeout to %d" % statement_timeout
                    )
                    cnx.commit()

                postgres_hooks = SQL_CONNECT_HOOKS["postgres"]
                postgres_hooks.append(set_postgres_timeout)

    def wrapped_connection(self):
        """open and return a connection to the database, wrapped into a class
        handling reconnection and all
        """
        return self.cnx_wrap(self)

    def get_connection(self):
        """open and return a connection to the database"""
        return self.dbhelper.get_connection()

    def _backup_restore_env(self):
        if (
            self.config["db-driver"] == "postgres"
            and self.config["db-password"] is not None
        ):
            return {"PGPASSWORD": self.config["db-password"]}

    def backup_to_file(self, backupfile, confirm):
        extra_env = self._backup_restore_env()
        for cmd in self.dbhelper.backup_commands(backupfile, keepownership=False):
            if _run_command(cmd, extra_env=extra_env):
                if not confirm("   [Failed] Continue anyway?", default="n"):
                    raise Exception(f"Failed command: {cmd}")

    def restore_from_file(self, backupfile, confirm, drop=True):
        extra_env = self._backup_restore_env()
        for cmd in self.dbhelper.restore_commands(
            backupfile, keepownership=False, drop=drop
        ):
            if _run_command(cmd, extra_env=extra_env):
                if not confirm("   [Failed] Continue anyway?", default="n"):
                    raise Exception(f"Failed command: {cmd}")

    def merge_args(self, args, query_args):
        if args is not None:
            newargs = {}
            for key, val in args.items():
                # convert cubicweb binary into db binary
                if isinstance(val, Binary):
                    val = self._binary(val.getvalue())
                # convert timestamp to utc.
                # expect SET TiME ZONE to UTC at connection opening time.
                # This shouldn't change anything for datetime without TZ.
                elif isinstance(val, datetime) and val.tzinfo is not None:
                    val = utcdatetime(val)
                elif isinstance(val, time) and val.tzinfo is not None:
                    val = utctime(val)
                newargs[key] = val
            # should not collide
            assert not (
                frozenset(newargs) & frozenset(query_args)
            ), f"unexpected collision: {frozenset(newargs) & frozenset(query_args)}"
            newargs.update(query_args)
            return newargs
        return query_args

    def process_result(self, cursor, cnx=None, column_callbacks=None):
        """return a list of CubicWeb compliant values from data in the given cursor"""
        return list(self.iter_process_result(cursor, cnx, column_callbacks))

    def iter_process_result(self, cursor, cnx, column_callbacks=None):
        """return a iterator on tuples of CubicWeb compliant values from data
        in the given cursor
        """
        # use two different implementations to avoid paying the price of
        # callback lookup for each *cell* in results when there is nothing to
        # lookup
        if not column_callbacks:
            return self.dbhelper.dbapi_module.process_cursor(
                cursor, self._dbencoding, Binary
            )
        assert cnx
        return self._cb_process_result(cursor, column_callbacks, cnx)

    def _cb_process_result(self, cursor, column_callbacks, cnx):
        # begin bind to locals for optimization
        descr = cursor.description
        encoding = self._dbencoding
        process_value = self._process_value
        binary = Binary
        # /end
        cursor.arraysize = 100
        while True:
            results = cursor.fetchmany()
            if not results:
                break
            for line in results:
                result = []
                for col, value in enumerate(line):
                    if value is None:
                        result.append(value)
                        continue
                    cbstack = column_callbacks.get(col, None)
                    if cbstack is None:
                        value = process_value(value, descr[col], encoding, binary)
                    else:
                        for cb in cbstack:
                            value = cb(self, cnx, value)
                    result.append(value)
                yield result

    def preprocess_entity(self, entity):
        """return a dictionary to use as extra argument to cursor.execute
        to insert/update an entity into a SQL database
        """
        attrs = {}
        eschema = entity.e_schema
        converters = getattr(self.dbhelper, "TYPE_CONVERTERS", {})
        for attr, value in entity.cw_edited.items():
            if value is not None and eschema.subject_relations[attr].final:
                atype = str(entity.e_schema.destination(attr))
                if atype in converters:
                    # It is easier to modify preprocess_entity rather
                    # than add_entity (native) as this behavior
                    # may also be used for update.
                    value = converters[atype](value)
                elif (
                    atype == "Password"
                ):  # XXX could be done using a TYPE_CONVERTERS callback
                    # if value is a Binary instance, this mean we got it
                    # from a query result and so it is already encrypted
                    if isinstance(value, Binary):
                        value = value.getvalue()
                    else:
                        value = crypt_password(value)
                    value = self._binary(value)
                elif isinstance(value, Binary):
                    value = self._binary(value.getvalue())
            attrs[SQL_PREFIX + str(attr)] = value
        attrs[SQL_PREFIX + "eid"] = entity.eid
        return attrs

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


set_log_methods(SQLAdapterMixIn, getLogger("cubicweb.sqladapter"))


# connection initialization functions ##########################################


def _install_sqlite_querier_patch():
    """This monkey-patch hotfixes a bug sqlite causing some dates to be returned as strings rather than
    date objects (http://www.sqlite.org/cvstrac/tktview?tn=1327,33)
    """
    from cubicweb.server.querier import QuerierHelper

    if hasattr(QuerierHelper, "_sqlite_patched"):
        return  # already monkey patched

    def wrap_execute(base_execute):
        def new_execute(*args, **kwargs):
            rset = base_execute(*args, **kwargs)
            if rset.description:
                found_date = False
                for row, rowdesc in zip(rset, rset.description):
                    for cellindex, (value, vtype) in enumerate(zip(row, rowdesc)):
                        if vtype in ("TZDatetime", "Date", "Datetime") and isinstance(
                            value, str
                        ):
                            found_date = True
                            value = value.rsplit(".", 1)[0]
                            try:
                                row[cellindex] = datetime.strptime(
                                    value, "%Y-%m-%d %H:%M:%S"
                                )
                            except Exception:
                                row[cellindex] = datetime.strptime(value, "%Y-%m-%d")
                            if vtype == "TZDatetime":
                                row[cellindex] = row[cellindex].replace(tzinfo=utc)
                        if vtype == "Time" and isinstance(value, str):
                            found_date = True
                            try:
                                row[cellindex] = datetime.strptime(value, "%H:%M:%S")
                            except Exception:
                                # DateTime used as Time?
                                row[cellindex] = datetime.strptime(
                                    value, "%Y-%m-%d %H:%M:%S"
                                )
                        if vtype == "Interval" and isinstance(value, int):
                            found_date = True
                            # XXX value is in number of seconds?
                            row[cellindex] = timedelta(0, value, 0)
                    if not found_date:
                        break
            return rset

        return new_execute

    QuerierHelper.execute = wrap_execute(QuerierHelper.execute)
    QuerierHelper._sqlite_patched = True


def _init_sqlite_connection(cnx):
    """Internal function that will be called to init a sqlite connection"""
    _install_sqlite_querier_patch()

    class group_concat:
        def __init__(self):
            self.values = set()

        def step(self, value):
            if value is not None:
                self.values.add(value)

        def finalize(self):
            return ", ".join(str(v) for v in self.values)

    cnx.create_aggregate("GROUP_CONCAT", 1, group_concat)

    def _limit_size(text, maxsize, format="text/plain"):
        if len(text) < maxsize:
            return text
        if format in ("text/html", "text/xhtml", "text/xml"):
            text = remove_html_tags(text)
        if len(text) > maxsize:
            text = text[:maxsize] + "..."
        return text

    def limit_size3(text, format, maxsize):
        return _limit_size(text, maxsize, format)

    cnx.create_function("LIMIT_SIZE", 3, limit_size3)

    def limit_size2(text, maxsize):
        return _limit_size(text, maxsize)

    cnx.create_function("TEXT_LIMIT_SIZE", 2, limit_size2)

    def weekday(ustr):
        try:
            dt = datetime.strptime(ustr, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            dt = datetime.strptime(ustr, "%Y-%m-%d")
        # expect sunday to be 1, saturday 7 while weekday method return 0 for
        # monday
        return (dt.weekday() + 1) % 7

    cnx.create_function("WEEKDAY", 1, weekday)

    cnx.cursor().execute("pragma foreign_keys = on")

    import yams.constraints

    yams.constraints.patch_sqlite_decimal()


sqlite_hooks = SQL_CONNECT_HOOKS.setdefault("sqlite", [])
sqlite_hooks.append(_init_sqlite_connection)


def _init_postgres_connection(cnx):
    """Internal function that will be called to init a postgresql connection"""
    cnx.cursor().execute("SET TIME ZONE UTC")
    # commit is needed, else setting are lost if the connection is first
    # rolled back
    cnx.commit()


postgres_hooks = SQL_CONNECT_HOOKS.setdefault("postgres", [])
postgres_hooks.append(_init_postgres_connection)



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unittest for cubicweb.dataimport.stores"""

from unittest.mock import patch, Mock
import datetime as DT

import pytz

from cubicweb.dataimport import stores
from cubicweb.devtools.testlib import CubicWebTC


class RQLObjectStoreTC(CubicWebTC):
    store_impl = stores.RQLObjectStore
    insert_group_attrs = dict(name="grp")
    insert_user_attrs = dict(login="lgn", upassword="pwd")
    source_name = "system"
    user_extid = None

    def test_base(self):
        with self.admin_access.repo_cnx() as cnx:
            store = self.store_impl(cnx)
            # Check data insertion
            group_eid = store.prepare_insert_entity(
                "CWGroup", **self.insert_group_attrs
            )
            user_eid = store.prepare_insert_entity("CWUser", **self.insert_user_attrs)
            store.prepare_insert_relation(user_eid, "in_group", group_eid)
            store.flush()
            store.commit()
            store.finish()
            user = cnx.execute('CWUser X WHERE X login "lgn"').one()
            self.assertEqual(user_eid, user.eid)
            self.assertTrue(user.creation_date)
            self.assertTrue(user.modification_date)
            self.assertTrue(user.cwuri)
            self.assertEqual(user.created_by[0].eid, cnx.user.eid)
            self.assertEqual(user.owned_by[0].eid, cnx.user.eid)
            self.assertEqual(user.cw_source[0].name, self.source_name)
            groups = cnx.execute('CWGroup X WHERE U in_group X, U login "lgn"')
            self.assertEqual(group_eid, groups.one().eid)
            # Check data update
            store.prepare_update_entity("CWGroup", group_eid, name="new_grp")
            store.flush()
            store.commit()
            store.finish()
            self.assertFalse(cnx.execute('CWGroup X WHERE X name "grp"'))
            self.assertTrue(cnx.execute('CWGroup X WHERE X name "new_grp"'))
            # Check data update with wrong type
            with self.assertRaises(AssertionError):
                store.prepare_update_entity("CWUser", group_eid, name="new_user")
            store.flush()
            store.commit()
            store.finish()
            self.assertFalse(cnx.execute('CWGroup X WHERE X name "new_user"'))
            self.assertTrue(cnx.execute('CWGroup X WHERE X name "new_grp"'))

    def test_store_rollback_on_error(self):
        """assert that the import is rolledback error"""

        class FakeException(Exception):
            pass

        with self.admin_access.repo_cnx() as cnx:
            cnx.rollback = Mock(wraps=cnx.rollback)
            try:
                with stores.RQLObjectStore(cnx) as store:
                    store.prepare_insert_entity("CWGroup", name="toto")
                    store.flush()
                    raise FakeException
            except FakeException:
                # something went wrong during the import, make sure the
                # transaction has been rolled back.
                cnx.rollback.assert_called_once()

            cnx.commit()
            # rollback has been called, but let's make sure that no cwgroups has
            # been created with the name "toto"
            rset = cnx.execute("Any X WHERE X is CWGroup, X name %(n)s", {"n": "toto"})
            self.assertEqual(len(rset), 0)

    @patch("cubicweb.dataimport.stores.RQLObjectStore.finish")
    @patch("cubicweb.dataimport.stores.RQLObjectStore.flush")
    def test_store_finish_called_on_flush_failure(self, flush, finish):
        """assert that the 'store.finish' and 'store.flush' are called if
        no exception is raised during import"""

        with self.admin_access.repo_cnx() as cnx:
            with stores.RQLObjectStore(cnx) as store:
                store.prepare_insert_entity("CWGroup", name="toto")

            flush.assert_called_once()
            finish.assert_called_once()


class NoHookRQLObjectStoreTC(RQLObjectStoreTC):
    store_impl = stores.NoHookRQLObjectStore


class NoHookRQLObjectStoreWithCustomMDGenStoreTC(RQLObjectStoreTC):
    insert_group_attrs = RQLObjectStoreTC.insert_group_attrs.copy()
    insert_group_attrs["cwuri"] = "http://somewhere.com/group/1"
    insert_user_attrs = RQLObjectStoreTC.insert_user_attrs.copy()
    insert_user_attrs["cwuri"] = "http://somewhere.com/user/1"
    source_name = "test"
    user_extid = b"http://somewhere.com/user/1"

    def store_impl(self, cnx):
        source = cnx.create_entity("CWSource", type="datafeed", name="test", url="test")
        cnx.commit()
        metagen = stores.MetadataGenerator(
            cnx, source=cnx.repo.source_by_eid(source.eid)
        )
        return stores.NoHookRQLObjectStore(cnx, metagen)


class MetaGeneratorTC(CubicWebTC):
    metagenerator_impl = stores.MetaGenerator
    _etype_rels = staticmethod(lambda x: x.etype_rels)

    def test_dont_generate_relation_to_internal_manager(self):
        with self.admin_access.repo_cnx() as cnx:
            metagen = self.metagenerator_impl(cnx)
            self.assertIn("created_by", self._etype_rels(metagen))
            self.assertIn("owned_by", self._etype_rels(metagen))
        with self.repo.internal_cnx() as cnx:
            metagen = self.metagenerator_impl(cnx)
            self.assertNotIn("created_by", self._etype_rels(metagen))
            self.assertNotIn("owned_by", self._etype_rels(metagen))

    def test_dont_generate_specified_values(self):
        with self.admin_access.repo_cnx() as cnx:
            metagen = self.metagenerator_impl(cnx)
            # hijack gen_modification_date to ensure we don't go through it
            metagen.gen_modification_date = None
            md = DT.datetime.now(pytz.utc) - DT.timedelta(days=1)
            entity, rels = metagen.base_etype_dicts("CWUser")
            entity.cw_edited.update(dict(modification_date=md))
            metagen.init_entity(entity)
            self.assertEqual(entity.cw_edited["modification_date"], md)


class MetadataGeneratorWrapperTC(MetaGeneratorTC):
    @staticmethod
    def metagenerator_impl(cnx):
        return stores._MetaGeneratorBWCompatWrapper(stores.MetadataGenerator(cnx))

    _etype_rels = staticmethod(lambda x: x._mdgen._etype_rels)


class MetadataGeneratorTC(CubicWebTC):
    def test_dont_generate_relation_to_internal_manager(self):
        with self.admin_access.repo_cnx() as cnx:
            metagen = stores.MetadataGenerator(cnx)
            self.assertIn("created_by", metagen.etype_rels("CWUser"))
            self.assertIn("owned_by", metagen.etype_rels("CWUser"))
        with self.repo.internal_cnx() as cnx:
            metagen = stores.MetadataGenerator(cnx)
            self.assertNotIn("created_by", metagen.etype_rels("CWUser"))
            self.assertNotIn("owned_by", metagen.etype_rels("CWUser"))

    def test_dont_generate_specified_values(self):
        with self.admin_access.repo_cnx() as cnx:
            metagen = stores.MetadataGenerator(cnx)
            # hijack gen_modification_date to ensure we don't go through it
            metagen.gen_modification_date = None
            md = DT.datetime.now(pytz.utc) - DT.timedelta(days=1)
            attrs = metagen.base_etype_attrs("CWUser")
            attrs.update(dict(modification_date=md))
            metagen.init_entity_attrs("CWUser", 1, attrs)
            self.assertEqual(attrs["modification_date"], md)


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""the Bookmark entity type for internal links

"""

from yams.buildobjs import EntityType, RelationType, SubjectRelation, String

from cubicweb import _
from cubicweb.schema import RRQLExpression


class Bookmark(EntityType):
    """bookmarks are used to have user's specific internal links"""

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (
            "managers",
            "users",
        ),
        "delete": (
            "managers",
            "owners",
        ),
        "update": (
            "managers",
            "owners",
        ),
    }

    title = String(required=True, maxsize=128, internationalizable=True)
    path = String(
        maxsize=2048,
        required=True,
        description=_("relative url of the bookmarked page"),
    )

    bookmarked_by = SubjectRelation(
        "CWUser", description=_("users using this bookmark")
    )


class bookmarked_by(RelationType):
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        # test user in users group to avoid granting permission to anonymous user
        "add": (
            "managers",
            RRQLExpression('O identity U, U in_group G, G name "users"'),
        ),
        "delete": (
            "managers",
            RRQLExpression('O identity U, U in_group G, G name "users"'),
        ),
    }



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""generate list of modules for sphinx doc

"""

EXCLUDE_DIRS = ("test", "tests", "examples", "data", "doc", "dist", ".hg", "migration")
if __name__ == "__main__":
    from logilab.common.sphinxutils import ModuleGenerator

    cw_gen = ModuleGenerator("cubicweb", "../..")
    cw_gen.generate(
        "../book/en/annexes/api_cubicweb.rst",
        EXCLUDE_DIRS + ("cwdesklets", "misc", "skel", "skeleton"),
    )
    for modname in ("logilab", "rql", "yams"):
        cw_gen = ModuleGenerator(modname, "../../../" + modname)
        cw_gen.generate(
            "../book/en/annexes/api_%s.rst" % modname, EXCLUDE_DIRS + ("tools",)
        )



from cubicweb.server.sources import datafeed


class SourceParserSuccess(datafeed.DataFeedParser):
    __regid__ = "test_source_parser_success"

    def process(self, url, raise_on_error=False):
        entity = self._cw.create_entity("Card", title="success")
        self.notify_updated(entity)


class SourceParserFail(SourceParserSuccess):
    __regid__ = "test_source_parser_fail"

    def process(self, url, raise_on_error=False):
        entity = self._cw.create_entity("Card", title="fail")
        self.notify_updated(entity)
        raise RuntimeError("fail")



#!/usr/bin/env python
# pylint: disable=W0142,W0403,W0404,W0613,W0622,W0622,W0704,R0904,C0103,E0611
#
# copyright 2003-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Generic Setup script, takes package info from __pkginfo__.py file
"""

from os.path import dirname, join

from setuptools import setup, find_packages

here = dirname(__file__)

# import required features
pkginfo = join(here, "cubicweb", "__pkginfo__.py")
__pkginfo__ = {}
with open(pkginfo) as f:
    exec(f.read(), __pkginfo__)
modname = __pkginfo__["modname"]
version = __pkginfo__["version"]
license = __pkginfo__["license"]
description = __pkginfo__["description"]
web = __pkginfo__["web"]
author = __pkginfo__["author"]
author_email = __pkginfo__["author_email"]

with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

# import optional features
distname = __pkginfo__["distname"]
package_data = __pkginfo__["package_data"]

setup(
    name=distname,
    version=version,
    license=license,
    url=web,
    description=description,
    long_description=long_description,
    author=author,
    author_email=author_email,
    packages=find_packages(),
    package_data=package_data,
    include_package_data=True,
    install_requires=[
        "logilab-common >= 1.8.3, < 2.0.0",
        "logilab-mtconverter >= 0.9.2, < 1.0.0",
        "rql >= 0.39, < 1.0.0",
        "yams >= 0.50.0, < 1.0.0",
        "lxml",
        "logilab-database >= 1.18.6, < 2.0.0",
        "passlib >= 1.7",
        "pytz",
        "Markdown >= 3.4.0",
        "filelock",
        "rdflib >= 6.0.0",
        "pyramid >= 1.9.0, < 2.0.0",
        "waitress >= 2.1.1, < 3.0.0",
        "wsgicors >= 0.3",
        "pyramid_multiauth < 1.0.0",  # to remove with Pyramid 2
        "repoze.lru",
        "cachetools",
        "importlib_metadata; python_version < '3.10'",
    ],
    entry_points={
        "console_scripts": [
            "cubicweb-ctl = cubicweb.cwctl:run",
        ],
        "paste.app_factory": [
            "pyramid_main=cubicweb.pyramid:pyramid_app",
        ],
    },
    extras_require={
        "captcha": [
            "Pillow",
        ],
        "crypto": [
            "pycryptodomex",
        ],
        "ext": [
            "docutils >= 0.6",
        ],
        "ical": [
            "vobject >= 0.6.0",
        ],
        "postgresql": [
            "psycopg2-binary",
        ],
        "s3": [
            "minio",
        ],
    },
    zip_safe=False,
    python_requires=">=3.7",
)
